#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a test suite chromosome."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pynguin.configuration as config
import pynguin.ga.chromosome as chrom
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.chromosomefactory as cf
    import pynguin.ga.chromosomevisitor as cv
    import pynguin.ga.testcasechromosome as tcc


class TestSuiteChromosome(chrom.Chromosome):
    """A chromosome that encodes a test suite."""

    def __init__(
        self,
        test_case_chromosome_factory: None
        | (cf.ChromosomeFactory[tcc.TestCaseChromosome]) = None,
        orig: TestSuiteChromosome | None = None,
    ):
        """
        Create new test suite chromosome.

        Args:
            test_case_chromosome_factory: Factory that produces new test case
                chromosomes. Required, if you want to mutated this
                chromosome.
            orig: Original, if we clone an existing chromosome.
        """
        super().__init__(orig=orig)

        if orig is None:
            self._test_case_chromosome_factory: None | (
                cf.ChromosomeFactory[tcc.TestCaseChromosome]
            ) = test_case_chromosome_factory
            self._test_case_chromosomes: list[tcc.TestCaseChromosome] = []
        else:
            self._test_case_chromosomes = [
                chromosome.clone() for chromosome in orig._test_case_chromosomes
            ]
            self._test_case_chromosome_factory = orig._test_case_chromosome_factory

    def add_test_case_chromosome(self, test: tcc.TestCaseChromosome) -> None:
        """Adds a test case chromosome to the test suite.

        Args:
            test: the test case to be added
        """
        self._test_case_chromosomes.append(test)
        self.set_changed(True)

    def delete_test_case_chromosome(self, test: tcc.TestCaseChromosome) -> None:
        """Delete a test case from the test suite.

        Args:
            test: the test case chromosome to delete
        """
        try:
            self._test_case_chromosomes.remove(test)
            self.set_changed(True)
        except ValueError:
            pass

    def add_test_case_chromosomes(self, tests: list[tcc.TestCaseChromosome]) -> None:
        """Adds a list of test cases to the test suite.

        Args:
            tests: A list of test case chromosomes to add
        """
        self._test_case_chromosomes.extend(tests)
        if tests:
            self.set_changed(True)

    def clone(self) -> TestSuiteChromosome:
        """Clones the chromosome.

        Returns:
            A clone of the chromosome  # noqa: DAR202
        """
        return TestSuiteChromosome(orig=self)

    def get_test_case_chromosome(self, index: int) -> tcc.TestCaseChromosome:
        """Provides the test case chromosome at a certain index.

        Args:
            index: the index to select

        Returns:
            The test case chromosome at the given index
        """
        return self._test_case_chromosomes[index]

    @property
    def test_case_chromosomes(self) -> list[tcc.TestCaseChromosome]:
        """Provides all test chromosomes.

        Returns:
            The list of all test cases
        """
        return self._test_case_chromosomes

    def set_test_case_chromosome(
        self, index: int, test: tcc.TestCaseChromosome
    ) -> None:
        """Sets a test chromosome at a certain index.

        Args:
            index: the index to set the chromosome
            test: the test case to set
        """
        self._test_case_chromosomes[index] = test
        self.set_changed(True)

    def size(self) -> int:
        """Provides the size of the chromosome, i.e., its number of test cases.

        Returns:
            The size of the chromosome
        """
        return len(self._test_case_chromosomes)

    def length(self) -> int:
        return sum(test.length() for test in self._test_case_chromosomes)

    def cross_over(
        self, other: chrom.Chromosome, position1: int, position2: int
    ) -> None:
        """Performs the crossover with another chromosome.

        Keep tests up to position1. Append copies of tests from other from position2
        onwards.

        Args:
            other: the other chromosome
            position1: the position in the first chromosome
            position2: the position in the second chromosome
        """
        assert isinstance(
            other, TestSuiteChromosome
        ), "Cannot perform crossover with " + str(type(other))

        self._test_case_chromosomes = self._test_case_chromosomes[:position1] + [
            test.clone() for test in other._test_case_chromosomes[position2:]
        ]
        self.set_changed(True)

    def mutate(self) -> None:
        """Apply mutation at test suite level."""
        assert (
            self._test_case_chromosome_factory is not None
        ), "Mutation is not possibly without test case chromosome factory"

        changed = False

        # Mutate existing test cases.
        for test in self._test_case_chromosomes:
            if randomness.next_float() < 1.0 / self.size():
                test.mutate()
                if test.has_changed():
                    changed = True

        # Randomly add new test cases.
        alpha = config.configuration.search_algorithm.test_insertion_probability
        exponent = 1
        while (
            randomness.next_float() <= pow(alpha, exponent)
            and self.size() < config.configuration.test_creation.max_size
        ):
            self.add_test_case_chromosome(
                self._test_case_chromosome_factory.get_chromosome()
            )
            exponent += 1
            changed = True

        # Remove any tests that have no more statements left.
        self._test_case_chromosomes = [
            t for t in self._test_case_chromosomes if t.size() > 0
        ]

        if changed:
            self.set_changed(True)

    def accept(self, visitor: cv.ChromosomeVisitor) -> None:
        visitor.visit_test_suite_chromosome(self)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, TestSuiteChromosome):
            return False
        if self.size() != other.size():
            return False
        for test, other_test in zip(
            self._test_case_chromosomes, other._test_case_chromosomes
        ):
            if test != other_test:
                return False
        return True

    def __hash__(self) -> int:
        return 31 + sum(17 * hash(t) for t in self._test_case_chromosomes)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provide various crossover functions for genetic algorithms."""
from abc import abstractmethod
from typing import Generic, TypeVar

import pynguin.ga.chromosome as chrom

# pylint: disable=invalid-name
T = TypeVar("T", bound=chrom.Chromosome)


# pylint: disable=too-few-public-methods
class CrossOverFunction(Generic[T]):
    """Cross over two individuals."""

    @abstractmethod
    def cross_over(self, parent1: T, parent2: T):
        """Perform a crossover between the two parents.

        Args:
            parent1: The first parent chromosome
            parent2: The second parent chromosome
        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides capabilities to create a test cluster"""
from __future__ import annotations

import dataclasses
import enum
import importlib
import inspect
import logging
import typing

from typing_inspect import get_args, is_union_type

import pynguin.configuration as config
from pynguin.setup.testcluster import FullTestCluster, TestCluster
from pynguin.typeinference import typeinference
from pynguin.typeinference.nonstrategy import NoTypeInferenceStrategy
from pynguin.typeinference.strategy import TypeInferenceStrategy
from pynguin.typeinference.stubstrategy import StubInferenceStrategy
from pynguin.typeinference.typehintsstrategy import TypeHintsInferenceStrategy
from pynguin.utils.exceptions import ConfigurationException
from pynguin.utils.generic.genericaccessibleobject import (
    GenericCallableAccessibleObject,
    GenericConstructor,
    GenericEnum,
    GenericFunction,
    GenericMethod,
)
from pynguin.utils.type_utils import (
    class_in_module,
    function_in_module,
    get_class_that_defined_method,
    is_primitive_type,
    is_type_unknown,
)


@dataclasses.dataclass(eq=True, frozen=True)
class DependencyPair:
    """
    Represents a dependency for a type that still needs to be resolved.
    We also store the recursion level, so we can enforce a limit on it.
    The recursion level is excluded from hash/eq so we don't get duplicate
    dependencies at different recursion levels.
    """

    dependency_type: type = dataclasses.field(compare=True, hash=True)
    recursion_level: int = dataclasses.field(compare=False, hash=False)


class TestClusterGenerator:  # pylint: disable=too-few-public-methods
    """Generate a new test cluster"""

    _logger = logging.getLogger(__name__)

    def __init__(self, modules_name: str):
        self._module_name = modules_name
        self._analyzed_classes: set[type] = set()
        self._dependencies_to_solve: set[DependencyPair] = set()
        self._test_cluster: FullTestCluster = FullTestCluster()
        self._inference = typeinference.TypeInference(
            strategies=self._initialise_type_inference_strategies()
        )
        self._enable_conditional_typing_imports()

    @staticmethod
    def _enable_conditional_typing_imports():
        """Enable imports that are conditional on the typing.TYPE_CHECKING variable.
        See https://docs.python.org/3/library/typing.html#typing.TYPE_CHECKING"""
        typing.TYPE_CHECKING = True

    @staticmethod
    def _initialise_type_inference_strategies() -> list[TypeInferenceStrategy]:
        strategy = config.configuration.type_inference.type_inference_strategy
        if strategy == config.TypeInferenceStrategy.NONE:
            return [NoTypeInferenceStrategy()]
        if strategy == config.TypeInferenceStrategy.STUB_FILES:
            if config.configuration.type_inference.stub_dir is None:
                raise ConfigurationException(
                    "Missing configuration value `stub_dir' for StubInferenceStrategy"
                )
            return [StubInferenceStrategy(config.configuration.type_inference.stub_dir)]
        if strategy == config.TypeInferenceStrategy.TYPE_HINTS:
            return [TypeHintsInferenceStrategy()]
        raise ConfigurationException("Invalid type-inference strategy")

    def generate_cluster(self) -> TestCluster:
        """Generate new test cluster from the configured module.

        Returns:
            The new test cluster
        """
        self._logger.debug("Generating test cluster")
        self._logger.debug("Analyzing module %s", self._module_name)
        module = importlib.import_module(self._module_name)
        for _, klass in inspect.getmembers(module, class_in_module(self._module_name)):
            self._add_dependency(klass, 1, True)

        for function_name, funktion in inspect.getmembers(
            module, function_in_module(self._module_name)
        ):

            generic_function = GenericFunction(
                funktion, self._inference.infer_type_info(funktion)[0], function_name
            )
            if self._is_protected(
                function_name
            ) or self._discard_accessible_with_missing_type_hints(generic_function):
                self._logger.debug("Skip function %s", function_name)
                continue

            self._logger.debug("Analyzing function %s", function_name)
            self._test_cluster.add_generator(generic_function)
            self._test_cluster.add_accessible_object_under_test(generic_function)
            self._add_callable_dependencies(generic_function, 1)
        self._resolve_dependencies_recursive()
        return self._test_cluster

    def _add_callable_dependencies(
        self, call: GenericCallableAccessibleObject, recursion_level: int
    ) -> None:
        """Add required dependencies.

        Args:
            call: The object whose parameter types should be added as dependencies.
            recursion_level: The current level of recursion of the search
        """
        self._logger.debug("Find dependencies for %s", call)
        if recursion_level > config.configuration.type_inference.max_cluster_recursion:
            self._logger.debug("Reached recursion limit. No more dependencies added.")
            return
        for param_name, type_ in call.inferred_signature.parameters.items():
            self._logger.debug("Resolving '%s' (%s)", param_name, type_)
            types = {type_}
            if is_union_type(type_):
                types = set(get_args(type_))

            for elem in types:
                if is_primitive_type(elem):
                    self._logger.debug("Not following primitive argument.")
                    continue
                if inspect.isclass(elem):
                    assert elem
                    self._logger.debug("Adding dependency for class %s", elem)
                    self._dependencies_to_solve.add(
                        DependencyPair(elem, recursion_level)
                    )
                else:
                    self._logger.debug("Found typing annotation %s, skipping", elem)
                    # TODO(fk) fully support typing annotations.

    def _add_dependency(self, klass: type, recursion_level: int, add_to_test: bool):
        """Add constructor/methods/attributes of the given type to the test cluster.

        Args:
            klass: The type of the dependency
            recursion_level: the current recursion level of the search
            add_to_test: whether the accessible objects are also added to objects
                under test.
        """
        assert inspect.isclass(klass), "Can only add dependencies for classes."
        if klass in self._analyzed_classes:
            self._logger.debug("Class %s already analyzed", klass)
            return
        self._analyzed_classes.add(klass)
        self._logger.debug("Analyzing class %s", klass)
        if issubclass(klass, enum.Enum):
            generic: GenericEnum | GenericConstructor = GenericEnum(klass)
        else:
            generic = generic_constructor = GenericConstructor(
                klass, self._inference.infer_type_info(klass.__init__)[0]
            )
            if self._discard_accessible_with_missing_type_hints(generic_constructor):
                return
            self._add_callable_dependencies(generic_constructor, recursion_level)

        self._test_cluster.add_generator(generic)
        if add_to_test:
            self._test_cluster.add_accessible_object_under_test(generic)

        for method_name, method in inspect.getmembers(klass, inspect.isfunction):
            # TODO(fk) why does inspect.ismethod not work here?!
            # TODO(fk) Instance methods of enums are only visible on elements of the
            # enum but not the class itself :|
            self._logger.debug("Analyzing method %s", method_name)

            generic_method = GenericMethod(
                klass, method, self._inference.infer_type_info(method)[0], method_name
            )

            if (
                self._is_constructor(method_name)
                or not self._is_method_defined_in_class(klass, method)
                or self._is_protected(method_name)
                or self._discard_accessible_with_missing_type_hints(generic_method)
            ):
                # Skip methods that should not be added to the cluster here.
                # Constructors are handled elsewhere; inherited methods should not be
                # part of the cluster, only overridden methods; private methods should
                # neither be part of the cluster.
                continue

            self._test_cluster.add_generator(generic_method)
            self._test_cluster.add_modifier(klass, generic_method)
            if add_to_test:
                self._test_cluster.add_accessible_object_under_test(generic_method)
            self._add_callable_dependencies(generic_method, recursion_level)
        # TODO(fk) how do we find attributes?

    @staticmethod
    def _is_constructor(method_name: str) -> bool:
        return method_name == "__init__"

    @staticmethod
    def _is_method_defined_in_class(class_: type, method: object) -> bool:
        return class_ == get_class_that_defined_method(method)

    @staticmethod
    def _is_protected(method_name: str) -> bool:
        return method_name.startswith("_") and not method_name.startswith("__")

    @staticmethod
    def _discard_accessible_with_missing_type_hints(
        accessible_object: GenericCallableAccessibleObject,
    ) -> bool:
        """Should we discard accessible objects that are not fully type hinted?

        Args:
            accessible_object: the object to check

        Returns:
            Whether or not the accessible should be discarded
        """
        if config.configuration.type_inference.guess_unknown_types:
            return False
        inf_sig = accessible_object.inferred_signature
        return any(
            is_type_unknown(type_) for param, type_ in inf_sig.parameters.items()
        )

    def _resolve_dependencies_recursive(self):
        """Resolve the currently open dependencies."""
        while self._dependencies_to_solve:
            to_solve = self._dependencies_to_solve.pop()
            self._add_dependency(
                to_solve.dependency_type, to_solve.recursion_level + 1, False
            )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a MIO."""
from __future__ import annotations

import logging
from dataclasses import dataclass
from math import ceil
from typing import TYPE_CHECKING

import pynguin.configuration as config
import pynguin.generation.algorithms.archive as arch
from pynguin.generation.algorithms.testgenerationstrategy import TestGenerationStrategy
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.testcasechromosome as tcc
    import pynguin.ga.testsuitechromosome as tsc


# pylint: disable=invalid-name
@dataclass
class Parameters:
    """Represents the parameters that are adjusted while running the algorithm."""

    # Probability for choosing creating a new test case or sampling an existing one.
    Pr: float = (
        config.configuration.mio.initial_config.random_test_or_from_archive_probability
    )

    # The maximum size of the population kept in the archive per target
    n: int = config.configuration.mio.initial_config.number_of_tests_per_target

    # The number of mutations performed on a test case before sampling again.
    m: int = config.configuration.mio.initial_config.number_of_mutations

    def is_valid(self):
        """Check if the parameters are valid."""
        assert self.Pr >= 0.0
        assert self.n >= 1
        assert self.m >= 1


# pylint: disable=too-few-public-methods
class MIOTestStrategy(TestGenerationStrategy[arch.MIOArchive]):
    """Implements MIO."""

    _logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self._solution: tcc.TestCaseChromosome | None = None
        self._parameters = Parameters()
        self._current_mutations = 0
        self._focused = False

    def generate_tests(
        self,
    ) -> tsc.TestSuiteChromosome:
        self.before_search_start()
        while (
            self.resources_left()
            and len(self._test_case_fitness_functions)
            - self._archive.num_covered_targets
            != 0
        ):
            self.evolve()
            self._update_parameters()
            self.after_search_iteration(self.create_test_suite(self._archive.solutions))
        self.after_search_finish()
        return self.create_test_suite(self._archive.solutions)

    # pylint:disable=line-too-long
    def _update_parameters(self):
        progress = self.progress()
        progress_until_focused = (
            progress / config.configuration.mio.exploitation_starts_at_percent
        )

        if self._focused:
            # Already in focused phase.
            # Nothing to change.
            return

        n_before = self._parameters.n
        if progress > config.configuration.mio.exploitation_starts_at_percent:
            self._logger.debug("Entering focused phase.")
            self._focused = True
            self._parameters.Pr = (
                config.configuration.mio.focused_config.random_test_or_from_archive_probability
            )
            self._parameters.n = (
                config.configuration.mio.focused_config.number_of_tests_per_target
            )
            self._parameters.m = (
                config.configuration.mio.focused_config.number_of_mutations
            )
        else:
            self._parameters.Pr = MIOTestStrategy._scale(
                config.configuration.mio.initial_config.random_test_or_from_archive_probability,
                config.configuration.mio.focused_config.random_test_or_from_archive_probability,
                progress_until_focused,
            )
            self._parameters.n = ceil(
                MIOTestStrategy._scale(
                    config.configuration.mio.initial_config.number_of_tests_per_target,
                    config.configuration.mio.focused_config.number_of_tests_per_target,
                    progress_until_focused,
                )
            )
            self._parameters.m = ceil(
                MIOTestStrategy._scale(
                    config.configuration.mio.initial_config.number_of_mutations,
                    config.configuration.mio.focused_config.number_of_mutations,
                    progress_until_focused,
                )
            )
        self._parameters.is_valid()
        if n_before != self._parameters.n:
            self._archive.shrink_solutions(self._parameters.n)

    @staticmethod
    def _scale(initial, focused, progress_until_focused):
        return initial + (focused - initial) * progress_until_focused

    def evolve(self) -> None:
        """Evolve the current population and replace it with a new one."""

        # From the second step on, MIO will decide to either sample a new test at random
        # (probability Pr), or will choose one existing test in the archive (probability
        # 1 - Pr), copy it, and mutate it.
        #
        # Note: in MIO there is an extra parameter m which controls how many mutations
        # and fitness evaluations should be done on the same individual before sampling
        # a new one.
        if self._solution is not None and self._current_mutations < self._parameters.m:
            offspring = self._solution.clone()
            offspring.mutate()
            self._current_mutations += 1
        elif randomness.next_float() < self._parameters.Pr:
            offspring = self.chromosome_factory.get_chromosome()
            self._current_mutations = 1
        else:
            maybe_offspring = self._archive.get_solution()
            if maybe_offspring is None:
                # Nothing in archive, so sample new one.
                offspring = self.chromosome_factory.get_chromosome()
            else:
                offspring = maybe_offspring
            offspring.mutate()
            self._current_mutations = 1
        if self._archive.update([offspring]):
            self._solution = offspring



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an abstract base class for chromosomes"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod

import pynguin.ga.chromosomevisitor as cv
import pynguin.ga.computations as ff


class Chromosome(metaclass=ABCMeta):
    """An abstract base class for chromosomes"""

    def __init__(self, orig: Chromosome | None = None):
        """
        Args:
            orig: Original, if we clone an existing chromosome.
        """
        if orig is None:
            self._computation_cache = ff.ComputationCache(self)
            self._changed: bool = True
            self._distance: float = -1
            self._rank: int = -1
        else:
            self._computation_cache = orig._computation_cache.clone(self)
            self._changed = orig._changed
            self._distance = orig._distance
            self._rank = orig._rank

    @abstractmethod
    def size(self) -> int:
        """Return the size of an individual.
        This should be number of elements it contains.

        Returns:
            The size of an individual  # noqa: DAR202
        """

    @abstractmethod
    def length(self) -> int:
        """Provide the length of an individual.
        This should be the total length of all contained elements and possible
        sub-elements. Look at the implementation to see the difference to size().

        Returns: The length of this individual.
        """

    def has_changed(self) -> bool:
        """Has this chromosome changed since the last evaluation?

        Returns:
            Whether or not this chromosome change since the last evaluation
        """
        return self._changed

    def set_changed(self, changed: bool) -> None:
        """Set changed status to parameter value.

        Args:
            changed: Then change state of this chromosome
        """
        self._changed = changed

    def get_fitness_functions(self) -> list[ff.FitnessFunction]:
        """Provide the currently configured fitness functions of this chromosome.

        Returns:
            The list of currently configured fitness functions
        """
        return self._computation_cache.get_fitness_functions()

    def add_fitness_function(
        self,
        fitness_function: ff.FitnessFunction,
    ) -> None:
        """Adds a fitness function.

        Args:
            fitness_function: A fitness function
        """
        self._computation_cache.add_fitness_function(fitness_function)

    def get_coverage_functions(self) -> list[ff.CoverageFunction]:
        """Provide the currently configured coverage functions of this chromosome.

        Returns:
            The list of currently configured coverage functions.
        """
        return self._computation_cache.get_coverage_functions()

    def add_coverage_function(
        self,
        coverage_function: ff.CoverageFunction,
    ) -> None:
        """Adds a coverage function.

        Args:
            coverage_function: A fitness function
        """
        self._computation_cache.add_coverage_function(coverage_function)

    def invalidate_cache(self) -> None:
        """Invalidate all cached computation values."""
        self._computation_cache.invalidate_cache()

    def get_fitness(self) -> float:
        """Provide a sum of the current fitness values

        Returns:
            The sum of the current fitness values
        """
        return self._computation_cache.get_fitness()

    def get_fitness_for(self, fitness_function: ff.FitnessFunction) -> float:
        """Returns the fitness values of a specific fitness function.

        Args:
            fitness_function: The fitness function

        Returns:
            Its fitness value
        """
        return self._computation_cache.get_fitness_for(fitness_function)

    def get_is_covered(self, fitness_function: ff.FitnessFunction) -> bool:
        """Check if the individual covers this fitness function.

        Args:
            fitness_function: The fitness function to check

        Returns:
            True, iff the individual covers the fitness function.
        """
        return self._computation_cache.get_is_covered(fitness_function)

    def get_coverage(self) -> float:
        """Provides the mean coverage value.

        Returns:
            The mean coverage value
        """
        return self._computation_cache.get_coverage()

    def get_coverage_for(self, coverage_function: ff.CoverageFunction) -> float:
        """Provides the coverage value for a certain coverage function

        Args:
            coverage_function: The fitness function who's coverage value shall be
                returned

        Returns:
            The coverage value for the fitness function
        """
        return self._computation_cache.get_coverage_for(coverage_function)

    @abstractmethod
    def cross_over(self, other: Chromosome, position1: int, position2: int) -> None:
        """Single point cross over.

        This chromosome will be split at `position1`, the other at `position2`,
        and the crossover will be performed with these pre- and suffixes.

        Args:
            other: The other chromosome to perform the crossover with
            position1: The point in the first chromosome
            position2: The point in the second chromosome
        """

    @abstractmethod
    def mutate(self) -> None:
        """Mutate this chromosome."""

    @abstractmethod
    def clone(self) -> Chromosome:
        """Create a clone of this chromosome.

        Returns:
            The cloned chromosome  # noqa: DAR202
        """

    @abstractmethod
    def accept(self, visitor: cv.ChromosomeVisitor) -> None:
        """Accept a chromosome visitor.

        Args:
            visitor: the visitor that is accepted.
        """

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def __hash__(self):
        pass

    @property
    def distance(self) -> float:
        """Provides the distance value of this chromosome.

        Returns:
            The distance value of this chromosome
        """
        return self._distance

    @distance.setter
    def distance(self, distance: float) -> None:
        """Sets the distance value of this chromosome.

        Args:
            distance: The new distance value
        """
        self._distance = distance

    @property
    def rank(self) -> int:
        """Provide the rank value of this chromosome.

        Returns:
            The rank value of this chromosome
        """
        return self._rank

    @rank.setter
    def rank(self, rank: int) -> None:
        """Sets the rank value of this chromosome.

        Args:
            rank: The new rank value
        """
        self._rank = rank



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""An exported implementation creating PyTest test cases from the statements."""
from __future__ import annotations

import ast
import os
from typing import TYPE_CHECKING

from pynguin.generation.export.abstractexporter import AbstractTestExporter

if TYPE_CHECKING:
    import pynguin.testcase.testcase as tc


# pylint: disable=too-few-public-methods
class PyTestExporter(AbstractTestExporter):
    """An exporter for PyTest-style test cases."""

    def export_sequences(self, path: str | os.PathLike, test_cases: list[tc.TestCase]):
        (
            module_aliases,
            common_modules,
            asts,
        ) = self._transform_to_asts(test_cases)
        import_nodes = AbstractTestExporter._create_ast_imports(
            module_aliases, common_modules
        )
        functions = AbstractTestExporter._create_functions(asts, False)
        module = ast.Module(body=import_nodes + functions)
        AbstractTestExporter._save_ast_to_file(path, module)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a data class to track the control-flow distances."""
from __future__ import annotations

from functools import total_ordering
from math import inf
from typing import TYPE_CHECKING, Any

import networkx as nx

import pynguin.ga.computations as ff

if TYPE_CHECKING:
    from pynguin.analyses.controlflow.controldependencegraph import (
        ControlDependenceGraph,
    )
    from pynguin.analyses.controlflow.programgraph import ProgramGraphNode
    from pynguin.testcase.execution import ExecutionResult, ExecutionTracer


@total_ordering
class ControlFlowDistance:
    """Tracks control-flow distances."""

    def __init__(self, approach_level: int = 0, branch_distance: float = 0.0) -> None:
        assert (
            approach_level >= 0 and branch_distance >= 0.0
        ), "Expect approach_level and branch_distance to be non-negative"
        self._approach_level = approach_level
        self._branch_distance = branch_distance

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, ControlFlowDistance):
            return False
        return (self._approach_level, self._branch_distance) == (
            other.approach_level,
            other.branch_distance,
        )

    def __lt__(self, other: ControlFlowDistance) -> bool:
        if not isinstance(other, ControlFlowDistance):
            raise TypeError(  # pylint: disable=raising-format-tuple
                "'<' not supported between instances of "
                "'ControlFlowDistance' and '%s'",
                type(other),
            )
        return (self._approach_level, self._branch_distance) < (
            other.approach_level,
            other.branch_distance,
        )

    @property
    def approach_level(self) -> int:
        """Provides the approach level.

        Returns:
            The approach level
        """
        return self._approach_level

    @approach_level.setter
    def approach_level(self, approach_level: int):
        assert approach_level >= 0, "Expect approach_level to be non-negative"
        self._approach_level = approach_level

    @property
    def branch_distance(self) -> float:
        """Provides the branch distance.

        Returns:
            The branch distance
        """
        return self._branch_distance

    @branch_distance.setter
    def branch_distance(self, branch_distance: float) -> None:
        assert branch_distance >= 0, "Expect branch_distance to be non-negative"
        self._branch_distance = branch_distance

    def increase_approach_level(self) -> None:
        """Increases the approach level by one."""
        self._approach_level += 1

    def get_resulting_branch_fitness(self) -> float:
        """Computes the resulting branch fitness.

        Returns:
            The resulting branch fitness
        """
        return self._approach_level + ff.normalise(self._branch_distance)

    def __str__(self) -> str:
        return (
            f"approach = {self._approach_level}, branch distance ="
            f" {self._branch_distance}"
        )

    def __repr__(self) -> str:
        return (
            f"ControlFlowDistance(approach_level={self._approach_level}, "
            f"branch_distance={self._branch_distance})"
        )


def get_root_control_flow_distance(
    result: ExecutionResult, code_object_id: int, tracer: ExecutionTracer
) -> ControlFlowDistance:
    """Computes the control flow distance for a root branch, i.e., if the given
    code object was executed.

    Args:
        result: the execution result.
        code_object_id: The code object id for which we want to get the root distance.
        tracer: the execution tracer

    Returns:
        The control flow distance, (0.0, 0.0) if it was executed, otherwise (1.0, 0.0)
    """
    assert code_object_id in tracer.get_known_data().branch_less_code_objects

    distance = ControlFlowDistance()
    if code_object_id in result.execution_trace.executed_code_objects:
        # The code object was executed by the execution
        return distance

    distance.increase_approach_level()
    return distance


def get_non_root_control_flow_distance(
    result: ExecutionResult, predicate_id: int, value: bool, tracer: ExecutionTracer
) -> ControlFlowDistance:
    """Computes the control flow distance for a predicate.

    Args:
        result: the execution result.
        predicate_id: The predicate id for which we want to get the root distance.
        value: compute distance to the true or the false branch?
        tracer: the execution tracer

    Returns:
        The control flow distance.
    """
    trace = result.execution_trace
    code_object_id = (
        tracer.get_known_data().existing_predicates[predicate_id].code_object_id
    )

    distance = ControlFlowDistance()
    # Code Object was not executed, simply use diameter as upper bound.
    if code_object_id not in trace.executed_code_objects:
        distance.approach_level = (
            tracer.get_known_data().existing_code_objects[code_object_id].cfg.diameter
        )
        return distance

    # Predicate was executed, simply use distance of correct branch.
    if predicate_id in trace.executed_predicates:
        if value:
            branch_distance = _predicate_fitness(predicate_id, trace.true_distances)
        else:
            branch_distance = _predicate_fitness(predicate_id, trace.false_distances)
        distance.branch_distance = branch_distance
        return distance

    cdg = tracer.get_known_data().existing_code_objects[code_object_id].cdg
    target_node = _get_node_with_predicate_id(cdg, predicate_id)

    # Choose diameter as upper bound
    distance.approach_level = (
        tracer.get_known_data().existing_code_objects[code_object_id].cfg.diameter
    )

    # We check for the closest predicate that was executed and compute the approach
    # level as the length of the path from such a predicate node to the desired
    # predicate node.
    for node in [
        node
        for node in cdg.nodes
        if node.predicate_id is not None
        and node.predicate_id in trace.executed_predicates
    ]:
        try:
            candidate = ControlFlowDistance()
            candidate.approach_level = nx.shortest_path_length(
                cdg.graph, node, target_node
            )
            # Predicate was executed but did not lead to execution of desired predicate
            # So the remaining branch distance to the true or false branch is
            # the desired distance, right?
            # One of them has to be zero, so we can simply add them.
            assert node.predicate_id is not None
            candidate.branch_distance = _predicate_fitness(
                node.predicate_id, trace.true_distances
            ) + _predicate_fitness(node.predicate_id, trace.false_distances)
            distance = min(distance, candidate)
        except nx.NetworkXNoPath:
            # No path from node to target.
            pass

    return distance


def _get_node_with_predicate_id(
    cdg: ControlDependenceGraph, predicate_id: int
) -> ProgramGraphNode:
    cdg_nodes = [node for node in cdg.nodes if node.predicate_id == predicate_id]
    assert len(cdg_nodes) == 1
    return cdg_nodes.pop()


def _predicate_fitness(predicate: int, branch_distances: dict[int, float]) -> float:
    return branch_distances.get(predicate, inf)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Pynguin is an automated unit test generation framework for Python.

This module provides the main entry location for the program executions.
"""
import sys

from pynguin.cli import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Utility methods for AST manipulation."""
from __future__ import annotations

import ast
import typing

if typing.TYPE_CHECKING:
    import pynguin.testcase.variablereference as vr
    import pynguin.utils.namingscope as ns


def create_full_name(
    variable_names: ns.AbstractNamingScope,
    module_names: ns.AbstractNamingScope,
    var: vr.Reference,
    load: bool,
) -> ast.Name | ast.Attribute:
    """Create a name node for the corresponding variable.

    Args:
        variable_names: the naming scope for the variables
        module_names: the naming scope for the modules
        var: the variable reference
        load: load or store?

    Returns:
        the name node
    """
    names = var.get_names(variable_names, module_names)
    loads = [True for _ in names[:-1]]
    loads.append(load)
    # First has to be ast.Name
    res: ast.Name | ast.Attribute = create_ast_name(names[0], not loads[0])
    # Remaining are ast.Attribute
    for name, loa in zip(names[1:], loads[1:]):
        res = create_ast_attribute(name, res, not loa)
    return res


def create_ast_name(name_id: str, store: bool = False) -> ast.Name:
    """Creates an AST name node.

    Args:
         name_id: for the name of the node
         store: store or load?

    Returns:
        the name node
    """
    return ast.Name(id=name_id, ctx=ast.Store() if store else ast.Load())


def create_ast_assign(target, value) -> ast.Assign:
    """Creates an AST assign node.

    Args:
         target: for the element where the value should be assigned to
         value: or the value which should be assigned

    Returns:
        the assign node
    """
    return ast.Assign(targets=[target], value=value)


def create_ast_attribute(attr, value, store: bool = False) -> ast.Attribute:
    """Creates an AST attribute node.

    Args:
        attr: for the name of the attribute
        value: for the value of the attribute
        store: store or load?

    Returns:
        the attribute node
    """
    return ast.Attribute(
        attr=attr, ctx=ast.Store() if store else ast.Load(), value=value
    )


def create_ast_list(elts, store: bool = False) -> ast.List:
    """Creates an AST list node.

    Args:
        elts: for the elements of the list
        store: store or load?

    Returns:
         the list node
    """
    return ast.List(ctx=ast.Store() if store else ast.Load(), elts=elts)


def create_ast_dict(keys, values) -> ast.Dict:
    """Creates an AST dict node.

    Args:
        keys: collection of the keys of the dictionary
        values: collection of the values of the dictionary

    Returns:
        the dict node
    """
    return ast.Dict(keys=keys, values=values)


def create_ast_set(elts) -> ast.Set:
    """Creates an AST set node.

    Args:
        elts: collection of the elements of the set

    Returns:
        the set node
    """
    return ast.Set(elts=elts)


def create_ast_tuple(elts, store: bool = False) -> ast.Tuple:
    """Creates an AST tuple node.

    Args:
        elts: collection of the elements of the tuple
        store: store or load?

    Returns:
        the tuple node
    """
    return ast.Tuple(ctx=ast.Store() if store else ast.Load(), elts=elts)


def create_ast_constant(value) -> ast.Constant:
    """Creates an AST constant node.

    Args:
        value: for the value of the constant

    Returns:
        the constant node
    """
    return ast.Constant(value=value, kind=None)


def create_ast_assert(test) -> ast.Assert:
    """Creates an AST assert node.

    Args:
        test: for the comparator

    Returns:
        the assert node
    """
    return ast.Assert(test=test, msg=None)


def create_ast_compare(left, operator, comparator) -> ast.Compare:
    """Creates an AST compare node.

    Args:
        left: for the element which should be compared
        operator: for the operator which should be used for comparison
        comparator: for the elements which should be used for comparison

    Returns:
        the compare node
    """
    return ast.Compare(left=left, ops=[operator], comparators=[comparator])


def create_ast_call(func, args, keywords) -> ast.Call:
    """Creates an AST call node.

    Args:
        func: for the callable which should be called
        args: collection of all the non positional arguments
        keywords: collection of all the positional arguments

    Returns:
        the call node
    """
    return ast.Call(func=func, args=args, keywords=keywords)


def create_ast_keyword(arg, value) -> ast.keyword:
    """Creates an AST keyword node.

    Args:
        arg: for the name of the keyword
        value: for the value of the keyword

    Returns:
        the keyword node
    """
    return ast.keyword(arg=arg, value=value)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides custom exception types."""


class ConfigurationException(BaseException):
    """An exception type that's raised if the generator has no proper configuration."""


class GenerationException(BaseException):
    """An exception during test generation.

    This type shall be used for all exceptions that occur during test generation and
    that are caused by the test-generation process.
    """


class ConstructionFailedException(BaseException):
    """An exception used when error occurs during construction of a test case."""


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides classes for computations on chromosomes, e.g., fitness and coverage."""
from __future__ import annotations

import abc
import dataclasses
import math
import statistics
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from pynguin.testcase.execution import ExecutionTrace

if TYPE_CHECKING:
    from pynguin.testcase.execution import ExecutionResult, KnownData, TestCaseExecutor


@dataclasses.dataclass(eq=False)
class ChromosomeComputation(abc.ABC):  # pylint:disable=too-few-public-methods
    """An abstract computation on chromosomes."""

    _executor: TestCaseExecutor
    """Executor that will be used by the computation to execute chromosomes."""


class TestCaseChromosomeComputation(
    ChromosomeComputation, metaclass=abc.ABCMeta
):  # pylint:disable=too-few-public-methods
    """A function that computes something on a test case chromosome."""

    def _run_test_case_chromosome(self, individual) -> ExecutionResult:
        """Runs a test suite and updates the execution results for
        all test cases that were changed.

        Args:
            individual: The individual to run

        Returns:
            A list of execution results
        """
        if individual.has_changed() or individual.get_last_execution_result() is None:
            individual.set_last_execution_result(
                self._executor.execute(individual.test_case)
            )
            individual.set_changed(False)
        result = individual.get_last_execution_result()
        assert result is not None
        return result


class TestSuiteChromosomeComputation(
    ChromosomeComputation, metaclass=abc.ABCMeta
):  # pylint:disable=too-few-public-methods
    """A function that computes something on a test suite chromosome."""

    def _run_test_suite_chromosome(self, individual) -> list[ExecutionResult]:
        """Runs a test suite and updates the execution results for
        all test cases that were changed.

        Args:
            individual: The individual to run

        Returns:
            A list of execution results
        """
        results: list[ExecutionResult] = []
        for test_case_chromosome in individual.test_case_chromosomes:
            if (
                test_case_chromosome.has_changed()
                or test_case_chromosome.get_last_execution_result() is None
            ):
                test_case_chromosome.set_last_execution_result(
                    self._executor.execute(test_case_chromosome.test_case)
                )
                test_case_chromosome.set_changed(False)
                # If we execute a suite which in turn executes it's test cases,
                # then we have to invalidate the values of the test cases, because
                # the test case is no longer aware that it was changed.
                test_case_chromosome.invalidate_cache()
            result = test_case_chromosome.get_last_execution_result()
            assert result is not None
            results.append(result)
        return results


class FitnessFunction:
    """Interface for a fitness function."""

    @abstractmethod
    def compute_fitness(self, individual) -> float:
        """Calculate the fitness value.

        Args:
            individual: the chromosome to compute the fitness for.

        Returns:
            the new fitness  # noqa: DAR202
        """

    @abstractmethod
    def compute_is_covered(self, individual) -> bool:
        """Compute if the goal of this fitness function is covered.
        This computation is usually cheaper than computing the fitness, because
        we are not interested in the distance, but only a boolean result.

        Args:
            individual: the chromosome to check coverage on.

        Returns:
            True, iff the goal of this fitness function is covered.
        """

    @abstractmethod
    def is_maximisation_function(self) -> bool:
        """Do we need to maximise or minimise this function?

        Returns:
             Whether or not this is a maximisation function  # noqa: DAR202
        """


class TestCaseFitnessFunction(
    TestCaseChromosomeComputation, FitnessFunction, metaclass=abc.ABCMeta
):
    """Base class for test case fitness functions."""

    def __init__(self, executor, code_object_id: int):
        super().__init__(executor)
        self._code_object_id = code_object_id

    @property
    def code_object_id(self) -> int:
        """The code object id, where the target of the fitness function is located.

        Returns:
            The code object id where the target of the fitness function is located.
        """
        return self._code_object_id


class BranchDistanceTestCaseFitnessFunction(TestCaseFitnessFunction):
    """A fitness function based on branch distances and entered code objects."""

    def compute_fitness(self, individual) -> float:
        result = self._run_test_case_chromosome(individual)
        merged_trace = analyze_results([result])
        tracer = self._executor.tracer

        return compute_branch_distance_fitness(merged_trace, tracer.get_known_data())

    def compute_is_covered(self, individual) -> bool:
        result = self._run_test_case_chromosome(individual)
        merged_trace = analyze_results([result])
        tracer = self._executor.tracer

        return compute_branch_distance_fitness_is_covered(
            merged_trace, tracer.get_known_data()
        )

    def is_maximisation_function(self) -> bool:
        return False


class TestSuiteFitnessFunction(
    TestSuiteChromosomeComputation, FitnessFunction, metaclass=abc.ABCMeta
):
    """Base class for test case fitness functions."""


class BranchDistanceTestSuiteFitnessFunction(TestSuiteFitnessFunction):
    """A fitness function based on branch distances and entered code objects."""

    def __init__(self, executor):
        super().__init__(executor)
        self._excluded_code_objects: set[int] = set()
        self._excluded_true_predicates: set[int] = set()
        self._excluded_false_predicates: set[int] = set()

    def restrict(
        self, exclude_code: set[int], exclude_true: set[int], exclude_false: set[int]
    ) -> None:
        """Restrict this fitness function, i.e., which branches/code objects it
        considers

        Args:
            exclude_code: Ids of the code objects that should not be considered.
            exclude_true: Ids of predicates whose True branch should not be considered.
            exclude_false: Ids of predicates whose False branch should not be
                considered.
        """
        self._excluded_code_objects.update(exclude_code)
        self._excluded_true_predicates.update(exclude_true)
        self._excluded_false_predicates.update(exclude_false)

    def compute_fitness(self, individual) -> float:
        results = self._run_test_suite_chromosome(individual)
        merged_trace = analyze_results(results)
        tracer = self._executor.tracer

        return compute_branch_distance_fitness(
            merged_trace,
            tracer.get_known_data(),
            self._excluded_code_objects,
            self._excluded_true_predicates,
            self._excluded_false_predicates,
        )

    def compute_is_covered(self, individual) -> bool:
        results = self._run_test_suite_chromosome(individual)
        merged_trace = analyze_results(results)
        tracer = self._executor.tracer

        return compute_branch_distance_fitness_is_covered(
            merged_trace,
            tracer.get_known_data(),
            self._excluded_code_objects,
            self._excluded_true_predicates,
            self._excluded_false_predicates,
        )

    def is_maximisation_function(self) -> bool:
        return False


class CoverageFunction:  # pylint:disable=too-few-public-methods
    """Interface for a coverage function."""

    @abstractmethod
    def compute_coverage(self, individual) -> float:
        """Compute the coverage of the given individual.

        Args:
            individual: the chromosome to compute the coverage for.

        Returns:
            The computed coverage.
        """


# pylint: disable=too-few-public-methods
class TestSuiteCoverageFunction(
    TestSuiteChromosomeComputation, CoverageFunction, metaclass=abc.ABCMeta
):
    """Base class for all coverage functions that act on test suite level."""


class TestCaseCoverageFunction(
    TestCaseChromosomeComputation, CoverageFunction, metaclass=abc.ABCMeta
):  # pylint: disable=too-few-public-methods
    """Base class for all coverage functions that act on test case level."""


class TestSuiteBranchCoverageFunction(TestSuiteCoverageFunction):
    """Computes branch coverage on test suites."""

    def compute_coverage(self, individual) -> float:
        results = self._run_test_suite_chromosome(individual)
        merged_trace = analyze_results(results)
        tracer = self._executor.tracer
        return compute_branch_coverage(merged_trace, tracer.get_known_data())


class TestCaseBranchCoverageFunction(TestCaseCoverageFunction):
    """Computes branch coverage on test cases."""

    def compute_coverage(self, individual) -> float:
        result = self._run_test_case_chromosome(individual)
        merged_trace = analyze_results([result])
        tracer = self._executor.tracer
        return compute_branch_coverage(merged_trace, tracer.get_known_data())


class ComputationCache:
    """Caches computation results and computes values on demand."""

    def __init__(
        self,
        chromosome,
        fitness_functions: list[FitnessFunction] | None = None,
        coverage_functions: list[CoverageFunction] | None = None,
        fitness_cache: dict[FitnessFunction, float] | None = None,
        is_covered_cache: dict[FitnessFunction, bool] | None = None,
        coverage_cache: dict[CoverageFunction, float] | None = None,
    ):  # pylint:disable=too-many-arguments
        self._chromosome = chromosome
        self._fitness_functions = fitness_functions if fitness_functions else []
        self._coverage_functions = coverage_functions if coverage_functions else []
        self._fitness_cache: dict[FitnessFunction, float] = (
            fitness_cache if fitness_cache else {}
        )
        self._is_covered_cache: dict[FitnessFunction, bool] = (
            is_covered_cache if is_covered_cache else {}
        )
        self._coverage_cache: dict[CoverageFunction, float] = (
            coverage_cache if coverage_cache else {}
        )

    def clone(self, new_chromosome) -> ComputationCache:
        """Create a deep copy of this cache.

        Args:
            new_chromosome: The chromosome with which this cache is associated.

        Returns:
            A deep copy.
        """
        return ComputationCache(
            new_chromosome,
            list(self._fitness_functions),
            list(self._coverage_functions),
            dict(self._fitness_cache),
            dict(self._is_covered_cache),
            dict(self._coverage_cache),
        )

    def get_fitness_functions(self) -> list[FitnessFunction]:
        """Provide the currently configured fitness functions of this chromosome.

        Returns:
            The list of currently configured fitness functions
        """
        return self._fitness_functions

    def add_fitness_function(
        self,
        fitness_function: FitnessFunction,
    ) -> None:
        """Adds the given fitness function.

        Args:
            fitness_function: A fitness function
        """
        assert (
            not fitness_function.is_maximisation_function()
        ), "Currently only minimization is supported"
        self._fitness_functions.append(fitness_function)

    def get_coverage_functions(self) -> list[CoverageFunction]:
        """Provide the currently configured coverage functions of this chromosome.

        Returns:
            The list of currently configured coverage functions.
        """
        return self._coverage_functions

    def add_coverage_function(
        self,
        coverage_function: CoverageFunction,
    ) -> None:
        """Adds a coverage function.

        Args:
            coverage_function: A fitness function
        """
        self._coverage_functions.append(coverage_function)

    T = TypeVar("T", CoverageFunction, FitnessFunction)

    def _check_cache(
        self,
        comp: Callable[[T | None], None],
        cache: dict[T, Any],
        funcs: list[T],
        only: T | None = None,
    ) -> None:
        """Check if values need to be computed.

        Args:
            comp: The function to execute, if values need to be computed.
            cache: The cache that should be checked.
            funcs: The functions that are used to fill the respective cache.
            only: Only compute the values for this function, optional.
        """
        if self._chromosome.has_changed():
            # If the chromosome has changed, we invalidate all values computed so far
            self.invalidate_cache()
            # Compute those values in which we are interested.
            comp(only)
            # Mark individual as no longer changed.
            self._chromosome.set_changed(False)
        elif len(cache) != len(funcs):
            # The individual has not changed, but not all values are cached.
            # So we might have to compute the missing ones.
            comp(only)

    def _compute_fitness(self, only: FitnessFunction | None = None):
        for fitness_func in self._fitness_functions if only is None else (only,):
            if fitness_func not in self._fitness_cache:
                new_value = fitness_func.compute_fitness(self._chromosome)
                assert (
                    not math.isnan(new_value)
                    and not math.isinf(new_value)
                    and new_value >= 0
                ), f"Invalid fitness value {new_value}"
                self._fitness_cache[fitness_func] = new_value
                # When computing a minimising fitness value, we can also determine
                # if the goal is covered, i.e. if it is zero.
                self._is_covered_cache[fitness_func] = new_value == 0.0

    def _compute_is_covered(self, only: FitnessFunction | None = None):
        for fitness_func in self._fitness_functions if only is None else (only,):
            if fitness_func not in self._is_covered_cache:
                new_value = fitness_func.compute_is_covered(self._chromosome)
                self._is_covered_cache[fitness_func] = new_value

    def _compute_coverage(self, only: CoverageFunction | None = None):
        for coverage_func in self._coverage_functions if only is None else (only,):
            if coverage_func not in self._coverage_cache:
                new_value = coverage_func.compute_coverage(self._chromosome)
                assert (
                    not math.isnan(new_value)
                    and not math.isinf(new_value)
                    and (0 <= new_value <= 1)
                ), f"Invalid coverage value {new_value}"
                self._coverage_cache[coverage_func] = new_value

    def invalidate_cache(self) -> None:
        """Invalidate all cached computation values."""
        self._fitness_cache.clear()
        self._is_covered_cache.clear()
        self._coverage_cache.clear()

    def get_fitness(self) -> float:
        """Provide a sum of the current fitness values

        Returns:
            The sum of the current fitness values
        """
        self._check_cache(
            self._compute_fitness,
            self._fitness_cache,
            self._fitness_functions,
        )
        return sum(self._fitness_cache.values())

    def get_fitness_for(self, fitness_function: FitnessFunction) -> float:
        """Returns the fitness values of a specific fitness function.

        Args:
            fitness_function: The fitness function

        Returns:
            Its fitness value
        """
        self._check_cache(
            self._compute_fitness,
            self._fitness_cache,
            self._fitness_functions,
            fitness_function,
        )
        return self._fitness_cache[fitness_function]

    def get_is_covered(self, fitness_function: FitnessFunction) -> bool:
        """Check if the individual covers this fitness function.

        Args:
            fitness_function: The fitness function to check

        Returns:
            True, iff the individual covers the fitness function.
        """
        self._check_cache(
            self._compute_is_covered,
            self._is_covered_cache,
            self._fitness_functions,
            fitness_function,
        )
        return self._is_covered_cache[fitness_function]

    def get_coverage(self) -> float:
        """Provides the mean coverage value.

        Returns:
            The mean coverage value
        """
        self._check_cache(
            self._compute_coverage,
            self._coverage_cache,
            self._coverage_functions,
        )
        return statistics.mean(self._coverage_cache.values())

    def get_coverage_for(self, coverage_function: CoverageFunction) -> float:
        """Provides the coverage value for a certain coverage function

        Args:
            coverage_function: The fitness function who's coverage value shall be
                returned

        Returns:
            The coverage value for the fitness function
        """
        self._check_cache(
            self._compute_coverage,
            self._coverage_cache,
            self._coverage_functions,
            coverage_function,
        )
        return self._coverage_cache[coverage_function]


def normalise(value: float) -> float:
    """Normalise a value.

    Args:
        value: The value to normalise

    Returns:
        The normalised value

    Raises:
        RuntimeError: if the value is negative
    """
    if value < 0:
        raise RuntimeError("Values to normalise cannot be negative")
    if math.isinf(value):
        return 1.0
    return value / (1.0 + value)


def analyze_results(results: list[ExecutionResult]) -> ExecutionTrace:
    """Merge the trace of the given results.

    Args:
        results: The list of execution results to analyze

    Returns:
        the merged traces.
    """
    merged = ExecutionTrace()
    for result in results:
        trace = result.execution_trace
        assert trace is not None
        merged.merge(trace)
    return merged


def compute_branch_distance_fitness(
    trace: ExecutionTrace,
    known_data: KnownData,
    exclude_code: set[int] | None = None,
    exclude_true: set[int] | None = None,
    exclude_false: set[int] | None = None,
) -> float:
    """Computes fitness based on covered branches and branch distances.

    Args:
        trace: The execution trace
        known_data: All known data
        exclude_code: Ids of the code objects that should not be considered.
        exclude_true: Ids of predicates whose True branch should not be considered.
        exclude_false: Ids of predicates whose False branch should not be considered.

    Returns:
        The computed fitness value
    """
    # Handle None. Cannot use empty set as default, because of mutable default args.
    exclude_code = set() if exclude_code is None else exclude_code

    # Check if all branch-less code objects were executed.
    code_objects_missing: float = len(
        known_data.branch_less_code_objects.difference(
            trace.executed_code_objects, exclude_code
        )
    )
    assert (
        code_objects_missing >= 0.0
    ), "Amount of non covered code objects cannot be negative"

    # Handle None for branches.
    exclude_true = set() if exclude_true is None else exclude_true
    exclude_false = set() if exclude_false is None else exclude_false

    # Check if all predicates are covered
    predicate_fitness: float = 0.0
    for predicate in known_data.existing_predicates:
        if predicate not in exclude_true:
            predicate_fitness += _predicate_fitness(
                predicate, trace.true_distances, trace
            )
        if predicate not in exclude_false:
            predicate_fitness += _predicate_fitness(
                predicate, trace.false_distances, trace
            )
    assert predicate_fitness >= 0.0, "Predicate fitness cannot be negative."
    total_fitness = code_objects_missing + predicate_fitness
    return total_fitness


def _predicate_fitness(
    predicate: int, branch_distances: dict[int, float], trace: ExecutionTrace
) -> float:
    if predicate in branch_distances and branch_distances[predicate] == 0.0:
        return 0.0
    if (
        predicate in trace.executed_predicates
        and trace.executed_predicates[predicate] >= 2
    ):
        return normalise(branch_distances[predicate])
    return 1.0


def compute_branch_distance_fitness_is_covered(
    trace: ExecutionTrace,
    known_data: KnownData,
    exclude_code: set[int] | None = None,
    exclude_true: set[int] | None = None,
    exclude_false: set[int] | None = None,
) -> bool:
    """Computes if all branches and code objects have been executed.

    Args:
        trace: The execution trace
        known_data: All known data
        exclude_code: Ids of the code objects that should not be considered.
        exclude_true: Ids of predicates whose True branch should not be considered.
        exclude_false: Ids of predicates whose False branch should not be considered.

    Returns:
        True, iff all branches were covered
    """
    # Handle None. Cannot use empty set as default, because of mutable default args.
    exclude_code = set() if exclude_code is None else exclude_code

    # Check if all branch-less code objects were executed.
    if (
        len(
            known_data.branch_less_code_objects.difference(
                trace.executed_code_objects, exclude_code
            )
        )
        > 0
    ):
        return False

    # Handle None for branches.
    exclude_true = set() if exclude_true is None else exclude_true
    exclude_false = set() if exclude_false is None else exclude_false

    # Check if all predicates are covered
    for predicate in known_data.existing_predicates:
        if (
            predicate not in exclude_true
            and (predicate, 0.0) not in trace.true_distances
        ):
            return False
        if (
            predicate not in exclude_false
            and (predicate, 0.0) not in trace.false_distances
        ):
            return False
    return True


def compute_branch_coverage(trace: ExecutionTrace, known_data: KnownData) -> float:
    """Computes branch coverage on bytecode instructions which should equal
    decision coverage on source.

    Args:
        trace: The execution trace
        known_data: All known data

    Returns:
        The computed coverage value
    """

    covered = len(
        trace.executed_code_objects.intersection(known_data.branch_less_code_objects)
    )
    existing = len(known_data.branch_less_code_objects)

    # Every predicate creates two branches
    existing += len(known_data.existing_predicates) * 2

    # A branch is covered if it has a distance of 0.0
    # Must consider both branches created by a predicate, i.e. true and false.
    covered += len([v for v in trace.true_distances.values() if v == 0.0])
    covered += len([v for v in trace.false_distances.values() if v == 0.0])

    if existing == 0:
        # Nothing to cover => everything is covered.
        coverage = 1.0
    else:
        coverage = covered / existing
    assert 0.0 <= coverage <= 1.0, "Coverage must be in [0,1]"
    return coverage


def compare(fitness_1: float, fitness_2: float) -> int:
    """Compare the two specified values.

    Args:
        fitness_1: The first value to compare
        fitness_2: The second value to compare

    Returns:
        the value 0 if fitness_1 is equal to fitness_2; a value less than 0 if
        fitness_1 is less than fitness_2; and a value greater than 0 if fitness_1 is
        greater than fitness_2
    """
    if fitness_1 < fitness_2:
        return -1
    if fitness_1 > fitness_2:
        return 1
    return 0



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a strategy that never does any type inference."""
from __future__ import annotations

import inspect
from typing import Callable

from pynguin.typeinference.strategy import InferredSignature, TypeInferenceStrategy


# pylint: disable=too-few-public-methods
class NoTypeInferenceStrategy(TypeInferenceStrategy):
    """Provides a strategy that never does any type inference."""

    def infer_type_info(self, method: Callable) -> InferredSignature:
        signature = inspect.signature(method)
        parameters: dict[str, type | None] = {}
        for param_name in signature.parameters:
            if param_name == "self":
                continue
            parameters[param_name] = None
        return_type: type | None = None

        return InferredSignature(
            signature=signature, parameters=parameters, return_type=return_type
        )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an assertion generator"""
from __future__ import annotations

import logging
from collections import Counter
from typing import TYPE_CHECKING

from ordered_set import OrderedSet

import pynguin.assertion.assertion as ass
import pynguin.assertion.assertiontraceobserver as ato
import pynguin.assertion.mutation_analysis.mutationadapter as ma
import pynguin.configuration as config
import pynguin.ga.chromosomevisitor as cv
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.testcasechromosome as tcc
    import pynguin.ga.testsuitechromosome as tsc
    import pynguin.testcase.execution as ex
    import pynguin.testcase.statement as st
    import pynguin.testcase.testcase as tc


class AssertionGenerator(cv.ChromosomeVisitor):
    """A simple assertion generator.
    Creates all regression assertions."""

    _logger = logging.getLogger(__name__)

    def __init__(self, executor: ex.TestCaseExecutor, base_executions: int = 2):
        """
        Create new assertion generator.

        Args:
            executor: the executor that will be used to execute the test cases.
            base_executions: How often should the tests be executed to filter
                out trivially flaky assertions, e.g., str representations based on
                memory locations.
        """
        self._base_executions = base_executions
        self._executor = executor
        self._executor.add_observer(ato.AssertionTraceObserver())

    def visit_test_suite_chromosome(self, chromosome: tsc.TestSuiteChromosome) -> None:
        self._generate_assertions(
            [chrom.test_case for chrom in chromosome.test_case_chromosomes]
        )

    def visit_test_case_chromosome(self, chromosome: tcc.TestCaseChromosome) -> None:
        self._generate_assertions([chromosome.test_case])

    def _generate_assertions(self, test_cases: list[tc.TestCase]) -> None:
        """Adds assertions to the given test case.

        Args:
            test_cases: the test case for which assertions should be generated.
        """
        self._add_assertions(self._execute(test_cases))

    def _execute(
        self, test_cases: list[tc.TestCase]
    ) -> list[tuple[tc.TestCase, list[ex.ExecutionResult]]]:
        tests_and_results: list[tuple[tc.TestCase, list[ex.ExecutionResult]]] = [
            (test_case, []) for test_case in test_cases
        ]
        for _ in range(self._base_executions):
            randomness.RNG.shuffle(tests_and_results)
            for test, results in tests_and_results:
                results.append(self._executor.execute(test))
        return tests_and_results

    def _add_assertions(
        self, tests_and_results: list[tuple[tc.TestCase, list[ex.ExecutionResult]]]
    ):
        for test_case, results in tests_and_results:
            self._add_assertions_for(test_case, results)

    def _add_assertions_for(
        self, test_case: tc.TestCase, results: list[ex.ExecutionResult]
    ):
        # In order to avoid repeating the same assertions after each statement,
        # we keep track of the last assertions and only assert things, if they
        # have changed.
        previous_statement_assertions = OrderedSet()
        for statement in test_case.statements:
            current_statement_assertions = self._get_assertions_for(results, statement)
            for assertion in current_statement_assertions:
                if (
                    not config.configuration.test_case_output.allow_stale_assertions
                    and assertion in previous_statement_assertions
                ):
                    # We already saw the same assertion in the previous statement
                    # So the value did not change.
                    continue
                if (
                    test_case.size_with_assertions()
                    >= config.configuration.test_case_output.max_length_test_case
                ):
                    self._logger.debug(
                        "No more assertions are added, because the maximum length "
                        "of a test case with its assertions was reached"
                    )
                    return
                statement.add_assertion(assertion)

            # Only update the previously seen assertions when we encounter a
            # statement that actually affects assertions.
            if statement.affects_assertions:
                previous_statement_assertions = current_statement_assertions

    def _get_assertions_for(  # pylint:disable=no-self-use
        self, results: list[ex.ExecutionResult], statement: st.Statement
    ) -> OrderedSet[ass.Assertion]:
        """Returns the set of assertions for the given statement.

        Args:
            results: The results from which we want to extract the assertions.
            statement: The statement after which the assertions hold.

        Returns:
            An ordered set of assertions for the given statement.
        """
        assert len(results) > 0, "Requires at least one result."
        assertions: list[OrderedSet[ass.Assertion]] = []
        for res in results:
            merged = OrderedSet()
            for trace in res.assertion_traces.values():
                merged.update(trace.get_assertions(statement))
            assertions.append(merged)

        first, *remainder = assertions
        first.intersection_update(*remainder)
        return first


class MutationAnalysisAssertionGenerator(AssertionGenerator):
    """Uses mutation analysis to filter out less relevant assertions."""

    def __init__(self, executor: ex.TestCaseExecutor):
        super().__init__(executor)
        adapter = ma.MutationAdapter()
        self._mutated_modules = [x for x, _ in adapter.mutate_module()]

    def _execute(
        self, test_cases: list[tc.TestCase]
    ) -> list[tuple[tc.TestCase, list[ex.ExecutionResult]]]:
        tests_and_results = super()._execute(test_cases)

        for idx, mutated_module in enumerate(self._mutated_modules):
            self._logger.info(
                "Running tests on mutant %3i/%i", idx + 1, len(self._mutated_modules)
            )
            self._executor.module_provider.add_mutated_version(
                module_name=config.configuration.module_name,
                mutated_module=mutated_module,
            )
            for test, results in tests_and_results:
                results.append(self._executor.execute(test))

        return tests_and_results

    def _get_assertions_for(
        self, results: list[ex.ExecutionResult], statement: st.Statement
    ) -> OrderedSet[ass.Assertion]:
        # The first executions are from executions on the non-mutated module.
        base_assertions = super()._get_assertions_for(
            results[: self._base_executions], statement
        )

        assertion_counter: dict[ass.Assertion, int] = Counter()
        for mutated_result in results[self._base_executions :]:
            for trace in mutated_result.assertion_traces.values():
                assertion_counter.update(Counter(trace.get_assertions(statement)))

        num_mutations = len(self._mutated_modules)

        # Only assertions that are not found every time are interesting.
        return OrderedSet(
            [
                assertion
                for assertion in base_assertions
                if assertion_counter[assertion] < num_mutations
            ]
        )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an abstract base class for a test generation algorithm."""
from __future__ import annotations

import time
from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, Iterable, TypeVar

from ordered_set import OrderedSet

import pynguin.ga.testsuitechromosome as tsc
import pynguin.generation.algorithms.archive as arch

if TYPE_CHECKING:
    import pynguin.coverage.branchgoals as bg
    import pynguin.ga.chromosomefactory as cf
    import pynguin.ga.computations as ff
    import pynguin.ga.testcasechromosome as tcc
    import pynguin.generation.searchobserver as so
    import pynguin.testcase.testfactory as tf
    from pynguin.ga.operators.crossover.crossover import CrossOverFunction
    from pynguin.ga.operators.ranking.rankingfunction import RankingFunction
    from pynguin.ga.operators.selection.selection import SelectionFunction
    from pynguin.generation.stoppingconditions.stoppingcondition import (
        StoppingCondition,
    )
    from pynguin.setup.testcluster import TestCluster
    from pynguin.testcase.execution import TestCaseExecutor

A = TypeVar("A", bound=arch.Archive)  # pylint:disable=invalid-name


class TestGenerationStrategy(
    Generic[A]
):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Provides an abstract base class for a test generation algorithm."""

    def __init__(self) -> None:
        self._archive: A
        self._chromosome_factory: cf.ChromosomeFactory
        self._executor: TestCaseExecutor
        self._test_cluster: TestCluster
        self._test_factory: tf.TestFactory
        self._selection_function: SelectionFunction
        self._stopping_condition: StoppingCondition
        self._crossover_function: CrossOverFunction
        self._ranking_function: RankingFunction
        self._test_case_fitness_functions: OrderedSet[
            ff.TestCaseFitnessFunction
        ] = OrderedSet()
        self._test_suite_fitness_functions: OrderedSet[
            ff.TestSuiteFitnessFunction
        ] = OrderedSet()
        self._test_suite_coverage_functions: OrderedSet[
            ff.TestSuiteCoverageFunction
        ] = OrderedSet()
        self._branch_goal_pool: bg.BranchGoalPool
        self._search_observers: list[so.SearchObserver] = []

    @property
    def chromosome_factory(self) -> cf.ChromosomeFactory:
        """Provides the chromosome factory.

        Returns:
            The chromosome factory
        """
        return self._chromosome_factory

    @chromosome_factory.setter
    def chromosome_factory(self, chromosome_factory):
        self._chromosome_factory = chromosome_factory

    @property
    def executor(self) -> TestCaseExecutor:
        """Provides the test-case executor

        Returns:
            The test-case executor
        """
        return self._executor

    @executor.setter
    def executor(self, executor: TestCaseExecutor) -> None:
        self._executor = executor

    @property
    def test_cluster(self) -> TestCluster:
        """Provide the test cluster.

        Returns:
            The test cluster
        """
        return self._test_cluster

    @test_cluster.setter
    def test_cluster(self, test_cluster: TestCluster) -> None:
        self._test_cluster = test_cluster

    @property
    def test_factory(self) -> tf.TestFactory:
        """Provide the test factory.

        Returns:
            The test factory
        """
        return self._test_factory

    @test_factory.setter
    def test_factory(self, test_factory: tf.TestFactory) -> None:
        self._test_factory = test_factory

    @property
    def archive(self) -> A:
        """Provide the used archive."""
        return self._archive

    @archive.setter
    def archive(self, archive: A) -> None:
        """Set the used archive."""
        self._archive = archive

    @property
    def branch_goal_pool(self) -> bg.BranchGoalPool:
        """Provides the used branch goal pool."""
        return self._branch_goal_pool

    @branch_goal_pool.setter
    def branch_goal_pool(self, branch_goal_pool: bg.BranchGoalPool) -> None:
        """Set the used branch goal."""
        self._branch_goal_pool = branch_goal_pool

    @property
    def selection_function(self) -> SelectionFunction:
        """Provides the used selection function.

        Returns:
            The used selection function
        """
        return self._selection_function

    @selection_function.setter
    def selection_function(self, selection_function: SelectionFunction) -> None:
        self._selection_function = selection_function

    @property
    def stopping_condition(self) -> StoppingCondition:
        """Provides the used stopping condition.

        Returns:
            The used stopping condition
        """
        return self._stopping_condition

    @stopping_condition.setter
    def stopping_condition(self, stopping_condition: StoppingCondition) -> None:
        self._stopping_condition = stopping_condition

    @property
    def crossover_function(self) -> CrossOverFunction:
        """Provides the used crossover function.

        Returns:
            The used crossover function
        """
        return self._crossover_function

    @crossover_function.setter
    def crossover_function(self, crossover_function: CrossOverFunction) -> None:
        self._crossover_function = crossover_function

    @property
    def ranking_function(self) -> RankingFunction:
        """Provides the used ranking function.

        Returns:
            The used ranking function
        """
        return self._ranking_function

    @ranking_function.setter
    def ranking_function(self, ranking_function: RankingFunction):
        self._ranking_function = ranking_function

    @property
    def test_case_fitness_functions(
        self,
    ) -> OrderedSet[ff.TestCaseFitnessFunction]:
        """Provides the list of test case fitness functions.

        Returns:
            The used test case fitness functions
        """
        return self._test_case_fitness_functions

    @test_case_fitness_functions.setter
    def test_case_fitness_functions(
        self,
        test_case_fitness_functions: OrderedSet[ff.TestCaseFitnessFunction],
    ) -> None:
        self._test_case_fitness_functions = test_case_fitness_functions

    @property
    def test_suite_fitness_functions(
        self,
    ) -> OrderedSet[ff.TestSuiteFitnessFunction]:
        """Provides the list of test suite fitness functions.

        Returns:
            The used test suite fitness functions
        """
        return self._test_suite_fitness_functions

    @test_suite_fitness_functions.setter
    def test_suite_fitness_functions(
        self,
        test_suite_fitness_functions: OrderedSet[ff.TestSuiteFitnessFunction],
    ) -> None:
        self._test_suite_fitness_functions = test_suite_fitness_functions

    @property
    def test_suite_coverage_functions(
        self,
    ) -> OrderedSet[ff.TestSuiteCoverageFunction]:
        """Provides the list of test suite coverage functions.

        Returns:
            The used test suite coverage functions
        """
        return self._test_suite_coverage_functions

    @test_suite_coverage_functions.setter
    def test_suite_coverage_functions(
        self,
        test_suite_coverage_functions: OrderedSet[ff.TestSuiteCoverageFunction],
    ) -> None:
        self._test_suite_coverage_functions = test_suite_coverage_functions

    def create_test_suite(
        self, population: Iterable[tcc.TestCaseChromosome]
    ) -> tsc.TestSuiteChromosome:
        """Wraps a population of test-case chromosomes in a test-suite chromosome.

        This will add the test suite fitness functions and coverage functions to
        the resulting chromosome.

        Args:
            population: A list of test-case chromosomes

        Returns:
            A test-suite chromosome
        """
        suite = tsc.TestSuiteChromosome()
        suite.add_test_case_chromosomes(list(population))
        for suite_fitness in self._test_suite_fitness_functions:
            suite.add_fitness_function(suite_fitness)
        for suite_coverage in self._test_suite_coverage_functions:
            suite.add_coverage_function(suite_coverage)
        return suite

    @abstractmethod
    def generate_tests(self) -> tsc.TestSuiteChromosome:
        """Generates tests for a given module until the time limit is reached.

        Returns:  # noqa: DAR202
            A chromosome containing all tests generated by the generation strategy.
        """

    def add_search_observer(self, observer: so.SearchObserver) -> None:
        """Add the given observer.

        Args:
            observer: The observer to add.
        """
        self._search_observers.append(observer)

    def before_search_start(self) -> None:
        """Has to be called when the search starts."""
        start = time.time_ns()
        for obs in self._search_observers:
            obs.before_search_start(start)

    def before_first_search_iteration(self, initial: tsc.TestSuiteChromosome) -> None:
        """Has to be called once before the very first iteration of the search
        algorithm. Calling this is optional, as not every approach has a result before
        the first iteration.

        Args:
            initial: The initially produced test suite.
        """
        for obs in self._search_observers:
            obs.before_first_search_iteration(initial)

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        """Has to be called after every iteration of the search algorithm.

        Args:
            best: The currently best produced test suite.
        """
        for obs in self._search_observers:
            obs.after_search_iteration(best)

    def after_search_finish(self) -> None:
        """Has to be called when the search has finished."""
        for obs in self._search_observers:
            obs.after_search_finish()

    def resources_left(self) -> bool:
        """Checks if there are still resources left, e.g., time or test case executions.

        Returns:
            Whether or not there are resources left.
        """
        return not self._stopping_condition.is_fulfilled()

    def progress(self) -> float:
        """Provides the progress of the search.

        Returns:
            A value in [0,1]."""
        limit = self._stopping_condition.limit()
        assert limit > 0.0
        return self._stopping_condition.current_value() / limit



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides capabilites to create a coverage report."""
from __future__ import annotations

import dataclasses
import datetime
import importlib.resources
import inspect
import sys
import typing
from pathlib import Path

import pygments
from jinja2 import Template
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.python import PythonLexer

import pynguin.configuration as config
import pynguin.ga.computations as ff

if typing.TYPE_CHECKING:
    import pynguin.ga.testsuitechromosome as tsc
    from pynguin.testcase.execution import TestCaseExecutor


@dataclasses.dataclass(frozen=True)
class CoverageEntry:
    """How many things exist and how many are covered?"""

    covered: int = 0
    existing: int = 0

    def __add__(self, other: CoverageEntry) -> CoverageEntry:
        """Add data from another coverage entry to this one.

        Args:
            other: another CoverageEntry whose values are added to this one.

        Returns:
            A new coverage entry with the summed up elements of self and other.
        """
        return CoverageEntry(
            self.covered + other.covered, self.existing + other.existing
        )


@dataclasses.dataclass
class LineAnnotation:
    """Coverage information for a single line."""

    line_no: int

    total: CoverageEntry

    branches: CoverageEntry

    branchless_code_objects: CoverageEntry

    def message(self):
        """Compute the message that should be displayed as a tool tip
        when hovering over this line number.

        Returns:
            The message for this line.
        """
        msgs = []
        if self.branches.existing > 0:
            msgs.append(
                f"{self.branches.covered}/{self.branches.existing} branches covered"
            )
        if self.branchless_code_objects.existing > 0:
            msgs.append(
                f"{self.branchless_code_objects.covered}/"
                f"{self.branchless_code_objects.existing}"
                f" branchless code objects covered"
            )
        return ";".join(msgs)


@dataclasses.dataclass
class CoverageReport:
    """All coverage related data required to create a coverage report."""

    module: str

    # Raw source code of the module under test
    source: list[str]

    # Achieved branch coverage
    branch_coverage: float

    # Information about total covered branches
    branches: CoverageEntry

    # Information about total covered branchless code objects
    branchless_code_objects: CoverageEntry

    line_annotations: list[LineAnnotation]


# pylint:disable=too-many-locals
def get_coverage_report(
    suite: tsc.TestSuiteChromosome, executor: TestCaseExecutor
) -> CoverageReport:
    """Create a coverage report for the given test suite

    Args:
        suite: The suite for which a coverage report should be generated.
        executor: The executor

    Returns:
        The coverage report.
    """
    results = []
    for test_case_chromosome in suite.test_case_chromosomes:
        result = test_case_chromosome.get_last_execution_result()
        assert result is not None
        results.append(result)
    trace = ff.analyze_results(results)
    known_data = executor.tracer.get_known_data()

    line_to_branchless_code_object_coverage = (
        _get_line_to_branchless_code_object_coverage(known_data, trace)
    )

    line_to_branch_coverage = _get_line_to_branch_coverage(known_data, trace)

    source = inspect.getsourcelines(sys.modules[config.configuration.module_name])[0]
    branch_coverage = ff.compute_branch_coverage(trace, known_data)

    branchless_code_objects = CoverageEntry()
    for cov in line_to_branchless_code_object_coverage.values():
        branchless_code_objects += cov
    branches = CoverageEntry()
    for cov in line_to_branch_coverage.values():
        branches += cov

    line_annotations = [
        _get_line_annotations(
            idx + 1, line_to_branchless_code_object_coverage, line_to_branch_coverage
        )
        for idx in range(len(source))
    ]

    return CoverageReport(
        config.configuration.module_name,
        source,
        branch_coverage,
        branches,
        branchless_code_objects,
        line_annotations,
    )


def render_coverage_report(
    cov_report: CoverageReport, report_path: Path, timestamp: datetime.datetime
) -> None:
    """Render the given coverage report to the given file.

    Args:
        timestamp: When was the report created.
        cov_report: The coverage report to render
        report_path: To file where the report should be rendered to.
    """
    with report_path.open(mode="w", encoding="utf-8") as html_file:
        template = Template(
            importlib.resources.read_text("pynguin.resources", "coverage-template.html")
        )
        html_file.write(
            template.render(
                cov_report=cov_report,
                highlight=pygments.highlight,
                lexer=PythonLexer,
                formatter=HtmlFormatter,
                date=timestamp,
            )
        )


def _get_line_to_branch_coverage(known_data, trace):
    line_to_branch_coverage = {}
    for predicate in known_data.existing_predicates:
        lineno = known_data.existing_predicates[predicate].line_no
        if lineno not in line_to_branch_coverage:
            line_to_branch_coverage[lineno] = CoverageEntry()
        line_to_branch_coverage[lineno] += CoverageEntry(existing=2)
        if (predicate, 0.0) in trace.true_distances.items():
            line_to_branch_coverage[lineno] += CoverageEntry(covered=1)
        if (predicate, 0.0) in trace.false_distances.items():
            line_to_branch_coverage[lineno] += CoverageEntry(covered=1)
    return line_to_branch_coverage


def _get_line_to_branchless_code_object_coverage(known_data, trace):
    line_to_branchless_code_object_coverage = {}
    for code in known_data.branch_less_code_objects:
        lineno = known_data.existing_code_objects[code].code_object.co_firstlineno
        if lineno not in line_to_branchless_code_object_coverage:
            line_to_branchless_code_object_coverage[lineno] = CoverageEntry()
        line_to_branchless_code_object_coverage[lineno] += CoverageEntry(existing=1)
        if code in trace.executed_code_objects:
            line_to_branchless_code_object_coverage[lineno] += CoverageEntry(covered=1)
    return line_to_branchless_code_object_coverage


def _get_line_annotations(
    lineno: int,
    code_object_coverage: dict[int, CoverageEntry],
    predicate_coverage: dict[int, CoverageEntry],
) -> LineAnnotation:
    """Compute line annotation for the given line no.

    Args:
        lineno: The lineno for which we should generate the information.
        code_object_coverage: code object coverage data
        predicate_coverage: predicate coverage data

    Returns:
        LineAnnotation data for the given line.
    """
    total = CoverageEntry()
    branches = CoverageEntry()
    branchless_code_objects = CoverageEntry()
    if lineno in code_object_coverage:
        branchless_code_objects = code_object_coverage[lineno]
        total += branchless_code_objects
    if lineno in predicate_coverage:
        branches = predicate_coverage[lineno]
        total += branches
    return LineAnnotation(lineno, total, branches, branchless_code_objects)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a comparator for preference sorting."""
from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

import pynguin.ga.chromosome as chrom

if TYPE_CHECKING:
    import pynguin.ga.computations as ff

C = TypeVar("C", bound=chrom.Chromosome)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class PreferenceSortingComparator(Generic[C]):
    """A comparator for chromosomes based on the fitness value of two objects.

    The comparator only considers the specified test goals.
    """

    def __init__(self, goal: ff.FitnessFunction) -> None:
        self._objective = goal

    # pylint: disable=too-many-return-statements
    def compare(self, solution_1: C | None, solution_2: C | None) -> int:
        """Compare the fitness value of two chromosome objects focusing only on one
        goal.

        Args:
            solution_1: A chromosome
            solution_2: A chromosome

        Returns:
            -1 if fitness of solution_1 is less than fitness of solution_2, 0 if the
            fitness values of both solutions are equal, 1 otherwise
        """
        if solution_1 is None:
            return 1
        if solution_2 is None:
            return -1

        value_1 = solution_1.get_fitness_for(self._objective)
        value_2 = solution_2.get_fitness_for(self._objective)
        if value_1 < value_2:
            return -1
        if value_1 > value_2:
            return 1
        if solution_1.length() < solution_2.length():
            return -1
        if solution_1.length() > solution_2.length():
            return 1
        return 0



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2020 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides some observers for statistics."""
from __future__ import annotations

import typing

import pynguin.generation.searchobserver as so
import pynguin.utils.statistics.statistics as stat
from pynguin.utils.statistics.runtimevariable import RuntimeVariable

if typing.TYPE_CHECKING:
    import pynguin.ga.testsuitechromosome as tsc


class IterationObserver(so.SearchObserver):
    """Observes the amount of iterations and logs them when the search
    has finished."""

    def __init__(self):
        self._iterations = 0

    def before_search_start(self, start_time_ns: int) -> None:
        pass

    def before_first_search_iteration(self, initial: tsc.TestSuiteChromosome) -> None:
        pass

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        self._iterations += 1

    def after_search_finish(self) -> None:
        stat.track_output_variable(
            RuntimeVariable.AlgorithmIterations, self._iterations
        )


class SequenceStartTimeObserver(so.SearchObserver):
    """Sets the start time for sequence bases statistics."""

    def before_search_start(self, start_time_ns: int) -> None:
        stat.set_sequence_start_time(start_time_ns)

    def before_first_search_iteration(self, initial: tsc.TestSuiteChromosome) -> None:
        pass

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        pass

    def after_search_finish(self) -> None:
        pass


class BestIndividualObserver(so.SearchObserver):
    """Observes the best individual."""

    def before_search_start(self, start_time_ns: int) -> None:
        pass

    def before_first_search_iteration(self, initial: tsc.TestSuiteChromosome) -> None:
        stat.current_individual(initial)

    def after_search_iteration(self, best: tsc.TestSuiteChromosome) -> None:
        stat.current_individual(best)

    def after_search_finish(self) -> None:
        pass



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an inference strategy for types."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from inspect import Parameter, Signature
from typing import Callable


@dataclass
class InferredSignature:
    """Encapsulates the types inferred for a method.

    The fields contain the following:
    - `signature`: Holds an `inspect.Signature` object as generated from the
      `inspect.signature` function
    - `parameters`: A dictionary mapping a parameter name to its type, if any.
    - `return_type`: The return type of a method, if any.

    The semantics of the `parameters` and `return_type` value for `None` is given as
    follows: the value `None` means that we do not yet know anything about this type,
    the value `NoneType` means that this parameter or return type is of type `None`,
    i.e., there is not parameter or return value.

    Consider the following example:
    - `def foo()` with `return_type = None` means we do not know what the return type is
    - `def bar() -> None` with `return_type = type(None) = NoneType` means that the
      function does not return anything.

    The types shall not be updated directly!  One is supposed to use the methods
    `update_parameter_type(parameter_name: str, parameter_type: Optional[type])` and
    `update_return_type(return_type: Optional[type])` to update the parameter or return
    type, respectively.  These methods will also adjust the value of the `signature`
    field by generating a new `inspect.Signature` instance accordingly.
    """

    signature: Signature
    parameters: dict[str, type | None] = field(default_factory=dict)
    return_type: type | None = None

    def update_parameter_type(
        self, parameter_name: str, parameter_type: type | None
    ) -> None:
        """Updates the type of one parameter.

        Args:
            parameter_name: The name of the parameter
            parameter_type: The new type of the parameter
        """
        assert parameter_name in self.parameters
        self.parameters[parameter_name] = parameter_type
        self._update_signature_parameter(parameter_name, parameter_type)

    def update_return_type(self, return_type: type | None) -> None:
        """Updates the return type

        Args:
            return_type: The new return type
        """
        self.return_type = return_type
        self._update_signature_return_type(return_type)

    def _update_signature_parameter(
        self,
        parameter_name: str,
        parameter_type: type | None,
    ):
        current_parameter: Parameter | None = self.signature.parameters.get(
            parameter_name
        )
        assert current_parameter is not None, "Cannot happen due to previous check"
        new_parameter = current_parameter.replace(annotation=parameter_type)
        new_parameters = [
            new_parameter if key == parameter_name else value
            for key, value in self.signature.parameters.items()
        ]
        new_signature = self.signature.replace(parameters=new_parameters)
        self.signature = new_signature

    def _update_signature_return_type(self, return_type: type | None):
        new_signature = self.signature.replace(return_annotation=return_type)
        self.signature = new_signature


# pylint: disable=too-few-public-methods
class TypeInferenceStrategy(metaclass=ABCMeta):
    """Provides an abstract base class for inference strategies for types."""

    @abstractmethod
    def infer_type_info(self, method: Callable) -> InferredSignature:
        """Infers the type information for a callable.

        Args:
            method: The callable we try to infer type information for

        Returns:
            A MethodType object with the inference results  # noqa: DAR202
        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a test cluster."""
from __future__ import annotations

import json
import logging
import typing
from abc import ABC, abstractmethod
from typing import Any

from ordered_set import OrderedSet
from typing_inspect import get_args, is_union_type

from pynguin.instrumentation.instrumentation import CODE_OBJECT_ID_KEY
from pynguin.utils import randomness, type_utils
from pynguin.utils.exceptions import ConstructionFailedException
from pynguin.utils.generic.genericaccessibleobject import (
    GenericCallableAccessibleObject,
)
from pynguin.utils.type_utils import COLLECTIONS, PRIMITIVES

if typing.TYPE_CHECKING:  # Break circular dependencies at runtime.
    import pynguin.ga.computations as ff
    import pynguin.generation.algorithms.archive as arch
    from pynguin.testcase.execution import KnownData
    from pynguin.utils.generic.genericaccessibleobject import GenericAccessibleObject


class TestCluster(ABC):
    """A test cluster which contains all methods/constructors/functions
    and all required transitive dependencies.
    """

    @property
    @abstractmethod
    def accessible_objects_under_test(self) -> OrderedSet[GenericAccessibleObject]:
        """Provides all accessible objects that are under test.

        Returns:
            The set of all accessible objects under test
        """

    @abstractmethod
    def num_accessible_objects_under_test(self) -> int:
        """Provide the number of accessible objects under test.

        This is useful to check if there even is something to test.

        Returns:
            The number of all accessibles under test
        """

    @abstractmethod
    def get_generators_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        """Retrieve all known generators for the given type.

        Args:
            for_type: The type we want to have the generators for

        Returns:
            The set of all generators for that type
        """

    @abstractmethod
    def get_modifiers_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        """Get all known modifiers of a type.

        This currently does not take inheritance into account.

        Args:
            for_type: The type

        Returns:
            The set of all accessibles that can modify the type
        """

    @property
    @abstractmethod
    def generators(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        """Provides all available generators.

        Returns:
            A dictionary of types and their generating accessibles
        """

    @property
    @abstractmethod
    def modifiers(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        """Provides all available modifiers.

        Returns:
            A dictionary of types and their modifying accessibles
        """

    @abstractmethod
    def get_random_accessible(self) -> GenericAccessibleObject | None:
        """Provide a random accessible of the unit under test.

        Returns:
            A random accessible
        """

    @abstractmethod
    def get_random_call_for(self, type_: type) -> GenericAccessibleObject:
        """Get a random modifier for the given type.

        Args:
            type_: The type

        Returns:
            A random modifier for that type

        Raises:
            ConstructionFailedException: if no modifiers for the type exist
        """

    @abstractmethod
    def get_all_generatable_types(self) -> list[type]:
        """Provides all types that can be generated, including primitives
        and collections.

        Returns:
            A list of all types that can be generated
        """

    @abstractmethod
    def select_concrete_type(self, select_from: type | None) -> type | None:
        """Select a concrete type from the given type.

        This is required e.g. when handling union types.
        Currently only unary types, Any and Union are handled.

        Args:
            select_from: An optional type

        Returns:
            An optional type
        """


class FullTestCluster(TestCluster):
    """A test cluster which contains all methods/constructors/functions
    and all required transitive dependencies.
    """

    def __init__(self):
        """Create new test cluster."""
        self._generators: dict[type, OrderedSet[GenericAccessibleObject]] = {}
        self._modifiers: dict[type, OrderedSet[GenericAccessibleObject]] = {}
        self._accessible_objects_under_test: OrderedSet[
            GenericAccessibleObject
        ] = OrderedSet()

    def add_generator(self, generator: GenericAccessibleObject) -> None:
        """Add the given accessible as a generator.

        It is only added if the type is known, not primitive and not NoneType.

        Args:
            generator: The accessible object
        """
        type_ = generator.generated_type()
        if (
            type_ is None
            or type_utils.is_none_type(type_)
            or type_utils.is_primitive_type(type_)
        ):
            return
        if type_ in self._generators:
            self._generators[type_].add(generator)
        else:
            self._generators[type_] = OrderedSet([generator])

    def add_accessible_object_under_test(self, obj: GenericAccessibleObject) -> None:
        """Add accessible object to the objects under test.

        Args:
            obj: The accessible object
        """
        self._accessible_objects_under_test.add(obj)

    def add_modifier(self, type_: type, obj: GenericAccessibleObject) -> None:
        """Add a modifier.

        A modified is something that can be used to modify the given type,
        e.g. a method.

        Args:
            type_: The type that can be modified
            obj: The accessible that can modify
        """
        if type_ in self._modifiers:
            self._modifiers[type_].add(obj)
        else:
            self._modifiers[type_] = OrderedSet([obj])

    @property
    def accessible_objects_under_test(self) -> OrderedSet[GenericAccessibleObject]:
        return self._accessible_objects_under_test

    def num_accessible_objects_under_test(self) -> int:
        return len(self._accessible_objects_under_test)

    def get_generators_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        if for_type in self._generators:
            return self._generators[for_type]
        return OrderedSet()

    def get_modifiers_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        if for_type in self._modifiers:
            return self._modifiers[for_type]
        return OrderedSet()

    @property
    def generators(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        return self._generators

    @property
    def modifiers(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        return self._modifiers

    def get_random_accessible(self) -> GenericAccessibleObject | None:
        if self.num_accessible_objects_under_test() == 0:
            return None
        return randomness.choice(self._accessible_objects_under_test)

    def get_random_call_for(self, type_: type) -> GenericAccessibleObject:
        accessible_objects = self.get_modifiers_for(type_)
        if len(accessible_objects) == 0:
            raise ConstructionFailedException("No modifiers for " + str(type_))
        return randomness.choice(accessible_objects)

    def get_all_generatable_types(self) -> list[type]:
        generatable = list(self._generators.keys())
        generatable.extend(PRIMITIVES)
        generatable.extend(COLLECTIONS)
        return generatable

    def select_concrete_type(self, select_from: type | None) -> type | None:
        if select_from == Any:  # pylint:disable=comparison-with-callable
            return randomness.choice(self.get_all_generatable_types())
        if is_union_type(select_from):
            possible_types = get_args(select_from)
            if possible_types is not None and len(possible_types) > 0:
                return randomness.choice(possible_types)
            return None
        return select_from


class FilteredTestCluster(TestCluster):
    """A test cluster that wraps another test cluster.
    This test cluster forwards most methods to the wrapped delegate.

    This test cluster filters out accessible objects under test that are already
    fully covered, in order to focus the search on areas that are not yet fully covered.
    """

    _logger = logging.getLogger(__name__)

    def __init__(
        self,
        test_cluster: TestCluster,
        archive: arch.Archive,
        known_data: KnownData,
        targets: OrderedSet[ff.TestCaseFitnessFunction],
    ):
        self._delegate = test_cluster
        self._known_data = known_data
        self._code_object_id_to_accessible_object: dict[
            int, GenericCallableAccessibleObject
        ] = {
            json.loads(acc.callable.__code__.co_consts[0])[CODE_OBJECT_ID_KEY]: acc
            for acc in test_cluster.accessible_objects_under_test
            if isinstance(acc, GenericCallableAccessibleObject)
            and hasattr(acc.callable, "__code__")
        }
        # Checking for __code__ is necessary, because the __init__ of a class that does
        # not define __init__ points to some internal CPython stuff.

        self._accessible_to_targets: dict[
            GenericCallableAccessibleObject, OrderedSet
        ] = {
            acc: OrderedSet()
            for acc in self._code_object_id_to_accessible_object.values()
        }
        for target in targets:
            if (acc := self._get_accessible_object_for_target(target)) is not None:
                targets_for_acc = self._accessible_to_targets.get(acc, None)
                targets_for_acc.add(target)

        # Get informed by archive, when a target is covered.
        archive.add_on_target_covered(self._on_target_covered)

    def _get_accessible_object_for_target(
        self, target: ff.TestCaseFitnessFunction
    ) -> GenericCallableAccessibleObject | None:
        code_object_id: int | None = target.code_object_id
        while code_object_id is not None:
            if (
                acc := self._code_object_id_to_accessible_object.get(
                    code_object_id, None
                )
            ) is not None:
                return acc
            code_object_id = self._known_data.existing_code_objects[
                code_object_id
            ].parent_code_object_id
        return None

    def _on_target_covered(self, target: ff.TestCaseFitnessFunction) -> None:
        acc = self._get_accessible_object_for_target(target)
        if acc is not None:
            targets_for_acc = self._accessible_to_targets.get(acc)
            assert targets_for_acc is not None
            targets_for_acc.remove(target)
            if len(targets_for_acc) == 0:
                self._accessible_to_targets.pop(acc)
                self._logger.debug(
                    "Removed %s from test cluster because all "
                    "targets within it are covered",
                    acc,
                )

    @property
    def accessible_objects_under_test(self) -> OrderedSet[GenericAccessibleObject]:
        accessibles = self._accessible_to_targets.keys()
        if len(accessibles) == 0:
            # Should never happen, just in case everything is already covered?
            return self._delegate.accessible_objects_under_test
        return OrderedSet(accessibles)

    def num_accessible_objects_under_test(self) -> int:
        return self._delegate.num_accessible_objects_under_test()

    def get_generators_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        return self._delegate.get_generators_for(for_type)

    def get_modifiers_for(self, for_type: type) -> OrderedSet[GenericAccessibleObject]:
        return self._delegate.get_modifiers_for(for_type)

    @property
    def generators(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        return self._delegate.generators

    @property
    def modifiers(self) -> dict[type, OrderedSet[GenericAccessibleObject]]:
        return self._delegate.modifiers

    def get_random_accessible(self) -> GenericAccessibleObject | None:
        accessibles = self._accessible_to_targets.keys()
        if len(accessibles) == 0:
            return self._delegate.get_random_accessible()
        return randomness.choice(OrderedSet(accessibles))

    def get_random_call_for(self, type_: type) -> GenericAccessibleObject:
        return self._delegate.get_random_call_for(type_)

    def get_all_generatable_types(self) -> list[type]:
        return self._delegate.get_all_generatable_types()

    def select_concrete_type(self, select_from: type | None) -> type | None:
        return self._delegate.select_concrete_type(select_from)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a comparator for dominance comparisons."""
from __future__ import annotations

from typing import Generic, TypeVar

from ordered_set import OrderedSet

import pynguin.ga.chromosome as chrom
import pynguin.ga.computations as ff

C = TypeVar("C", bound=chrom.Chromosome)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class DominanceComparator(Generic[C]):
    """Implements a comparator to compare chromosomes based on the dominance test."""

    def __init__(
        self,
        *,
        goal: ff.FitnessFunction | None = None,
        goals: OrderedSet[ff.FitnessFunction] | None = None,
    ) -> None:
        if goals is not None:
            self._objectives: OrderedSet[ff.FitnessFunction] | None = goals
        elif goal is not None:
            self._objectives = {goal}
        else:
            self._objectives = None

    # pylint: disable=too-many-return-statements
    def compare(self, chromosome_1: C | None, chromosome_2: C | None) -> int:
        """Compares two chromosomes regarding their dominance.

        Args:
            chromosome_1: The first chromosome
            chromosome_2: The second chromosome

        Returns:
            -1 if chromosome_1 dominates chromosome_2; 1 if chromosome_1 is dominated by
            chromosome_2; 0 otherwise
        """
        if chromosome_1 is None:
            return 1
        if chromosome_2 is None:
            return -1

        dominate_1 = False
        dominate_2 = False

        if self._objectives is None:
            self._objectives = set(chromosome_1.get_fitness_functions())

        for objective in self._objectives:
            flag = ff.compare(
                chromosome_1.get_fitness_for(objective),
                chromosome_2.get_fitness_for(objective),
            )
            if flag < 0:
                dominate_1 = True
                if dominate_2:
                    return 0
            elif flag > 0:
                dominate_2 = True
                if dominate_1:
                    return 0

        if dominate_1 == dominate_2:
            return 0  # no one dominates the other
        if dominate_1:
            return -1  # chromosome_1 dominates
        return 1  # chromosome_2 dominates



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a chromosome visitor"""

from abc import ABC, abstractmethod


class ChromosomeVisitor(ABC):
    """An abstract chromosome visitor."""

    @abstractmethod
    def visit_test_suite_chromosome(self, chromosome) -> None:
        """Visit a test suite chromosome.

        Args:
            chromosome: The test suite chromosome
        """

    @abstractmethod
    def visit_test_case_chromosome(self, chromosome) -> None:
        """Visit a test suite chromosome.

        Args:
            chromosome: The test case chromosome
        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a default implementation of a test case."""
from __future__ import annotations

import logging
from itertools import islice
from typing import TYPE_CHECKING, Any

from ordered_set import OrderedSet

import pynguin.testcase.testcase as tc

if TYPE_CHECKING:
    import pynguin.assertion.assertion as ass
    import pynguin.testcase.statement as stmt
    import pynguin.testcase.testcasevisitor as tcv
    import pynguin.testcase.variablereference as vr


# pylint:disable=too-many-public-methods
class DefaultTestCase(tc.TestCase):
    """A default implementation of a test case."""

    _logger = logging.getLogger(__name__)

    # pylint: disable=invalid-name
    def __init__(self) -> None:
        super().__init__()
        self._id = self._id_generator.inc()

    @property
    def id(self) -> int:
        """Get an unique ID representing this test case.

        Mainly useful for debugging.

        Returns:
            An unique ID representing this test case
        """
        return self._id

    def accept(self, visitor: tcv.TestCaseVisitor) -> None:
        visitor.visit_default_test_case(self)

    def add_statement(
        self, statement: stmt.Statement, position: int = -1
    ) -> vr.VariableReference | None:
        if position == -1:
            self._statements.append(statement)
        else:
            self._statements.insert(position, statement)
        return statement.ret_val

    def add_variable_creating_statement(
        self, statement: stmt.VariableCreatingStatement, position: int = -1
    ) -> vr.VariableReference:
        if position == -1:
            self._statements.append(statement)
        else:
            self._statements.insert(position, statement)
        return statement.ret_val

    def add_statements(self, statements: list[stmt.Statement]) -> None:
        self._statements.extend(statements)

    def append_test_case(self, test_case: tc.TestCase) -> None:
        memo: dict[vr.VariableReference, vr.VariableReference] = {}
        for statement in test_case.statements:
            clone = statement.clone(self, memo)
            if statement.ret_val is not None:
                # If the original statement created a variable, then so does the clone
                # Thus we know that clone.ret_val is not None
                memo[statement.ret_val] = clone.ret_val  # type: ignore
            self._statements.append(clone)

    def remove(self, position: int) -> None:
        self._logger.debug("Removing statement at position %d", position)
        if position >= self.size():
            return
        del self._statements[position]

    def remove_statement(self, statement: stmt.Statement) -> None:
        self._statements.remove(statement)

    def chop(self, pos: int) -> None:
        assert pos >= 0
        while len(self._statements) > pos + 1:
            del self._statements[-1]

    def contains(self, statement: stmt.Statement) -> bool:
        return statement in self._statements

    def get_statement(self, position: int) -> stmt.Statement:
        assert 0 <= position < len(self._statements)
        return self._statements[position]

    def set_statement(
        self, statement: stmt.Statement, position: int
    ) -> vr.VariableReference | None:
        assert 0 <= position < len(self._statements)
        self._statements[position] = statement
        return statement.ret_val

    def has_statement(self, position: int) -> bool:
        return 0 <= position < len(self._statements)

    def clone(self, limit: int | None = None) -> tc.TestCase:
        test_case = DefaultTestCase()
        memo: dict[vr.VariableReference, vr.VariableReference] = {}
        for statement in islice(self._statements, limit):
            copy = statement.clone(test_case, memo)
            if statement.ret_val is not None:
                # If the original statement created a variable, then so does the clone
                # Thus we know that clone.ret_val is not None
                memo[statement.ret_val] = copy.ret_val  # type: ignore
            test_case._statements.append(copy)
            copy.assertions = statement.copy_assertions(memo)
        test_case._id = self._id_generator.inc()
        return test_case

    def get_dependencies(
        self, var: vr.VariableReference
    ) -> OrderedSet[vr.VariableReference]:
        dependencies = OrderedSet()

        # TODO(fk) a variable will be a dependency of itself?!
        dependent_stmts = {self.get_statement(var.get_statement_position())}
        for idx in range(var.get_statement_position(), -1, -1):
            new_stmts = OrderedSet()
            for statement in dependent_stmts:
                if (
                    ret_val := self.get_statement(idx).ret_val
                ) is not None and statement.references(ret_val):
                    new_stmts.add(self.get_statement(idx))
                    dependencies.add(ret_val)
                    break
            dependent_stmts.update(new_stmts)

        return dependencies

    def get_assertions(self) -> list[ass.Assertion]:
        assertions: list[ass.Assertion] = []
        for statement in self._statements:
            assertions.extend(statement.assertions)
        return assertions

    def size_with_assertions(self) -> int:
        return self.size() + len(self.get_assertions())

    def size(self) -> int:
        return len(self._statements)

    # pylint: disable=too-many-return-statements
    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, DefaultTestCase):
            return False

        if not self._statements:
            if other._statements:
                return False
        else:
            if len(self._statements) != len(other._statements):
                return False
            memo: dict[vr.VariableReference, vr.VariableReference] = {}
            for left, right in zip(self._statements, other._statements):
                if ((lret := left.ret_val) is None) ^ ((rret := right.ret_val) is None):
                    # One is None but the other isn't, i.e., one creates a variable
                    # but the other doesn't -> they are different.
                    return False
                if lret is not None:
                    # lret is not None, so rret is also not None, otherwise we would be
                    # in the above case.
                    memo[lret] = rret  # type:ignore
                if not left.structural_eq(right, memo):
                    return False
        return True

    def __hash__(self) -> int:
        return 31 + sum(17 * s.structural_hash() for s in self._statements)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a factory to create test suite chromosomes."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ordered_set import OrderedSet

import pynguin.configuration as config
import pynguin.ga.chromosomefactory as cf
import pynguin.ga.testsuitechromosome as tsc
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.ga.computations as ff


class TestSuiteChromosomeFactory(
    cf.ChromosomeFactory[tsc.TestSuiteChromosome]
):  # pylint:disable=too-few-public-methods.
    """A factory that provides new test suite chromosomes of random length."""

    def __init__(
        self,
        test_case_chromosome_factory: cf.ChromosomeFactory,
        fitness_functions: OrderedSet[ff.TestSuiteFitnessFunction],
        coverage_functions: OrderedSet[ff.TestSuiteCoverageFunction],
    ):
        """Instantiates a new factory

        Args:
            test_case_chromosome_factory: The internal test case chromosome factory,
                                          which provides the test case chromosomes that
                                          will be part of a newly generated test suite
                                          chromosome
            fitness_functions: The fitness functions that will be added to every
                               newly generated chromosome.
            coverage_functions: The coverage functions that will be added to every
                                newly generated chromosome.
        """
        self._test_case_chromosome_factory = test_case_chromosome_factory
        self._fitness_functions = fitness_functions
        self._coverage_functions = coverage_functions

    def get_chromosome(self) -> tsc.TestSuiteChromosome:
        chromosome = tsc.TestSuiteChromosome(self._test_case_chromosome_factory)
        num_tests = randomness.next_int(
            config.configuration.search_algorithm.min_initial_tests,
            config.configuration.search_algorithm.max_initial_tests + 1,
        )

        for _ in range(num_tests):
            chromosome.add_test_case_chromosome(
                self._test_case_chromosome_factory.get_chromosome()
            )
        for func in self._fitness_functions:
            chromosome.add_fitness_function(func)
        for func in self._coverage_functions:
            chromosome.add_coverage_function(func)
        return chromosome



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a singleton instance of Random that can be seeded."""
from __future__ import annotations

import random
import string
from typing import Any, Sequence, TypeVar


class Random(random.Random):
    """Override Random to allow querying for the seed value.

    It generates a seed if none was given from `time.time_ns()`.  This is NOT
    cryptographically safe, and this random-number generator should not be used for
    anything related to cryptography.  For our case, however, it is good enough to
    use the current time stamp in nano seconds as seed.
    """

    def __init__(self, x=None) -> None:
        super().__init__(x)
        self._current_seed: int | None = None
        self.seed(x)

    # pylint: disable=import-outside-toplevel
    def seed(self, a=None, version: int = 2) -> None:
        if a is None:
            import time

            a = time.time_ns()

        self._current_seed = a
        super().seed(a)

    def get_seed(self) -> int:
        """Provides the used seed for random-number generation.

        Returns:
            Provides the used seed
        """
        assert self._current_seed is not None
        return self._current_seed


RNG: Random = Random()
RNG.seed()


def next_char() -> str:
    """Create a random printable ascii char.

    Returns:
        A random printable ascii char
    """
    return RNG.choice(string.printable)


def next_string(length: int) -> str:
    """Create a random string consisting of printable and with the given length.

    Args:
        length: the desired length

    Returns:
        A string of given length
    """
    return "".join(next_char() for _ in range(length))


def next_int(lower_bound=-100, upper_bound=100) -> int:
    """Provide a random integer number from an interval.

    If no lower or upper bound is given, the integer is chosen from the interval
    including -100 to excluded 100.

    Args:
        lower_bound: The lower bound for the number selection,
        upper_bound: The upper bound for the number selection, excluded

    Returns:
        A random integer from the interval
    """
    return RNG.randrange(lower_bound, upper_bound)


def next_float(lower_bound=0, upper_bound=1) -> float:
    """Provide a random float number uniformly selected from an interval.

    If no lower or upper bound is given, the float is chosen uniformly from the
    interval [0,1].

    Args:
        lower_bound: The lower bound for the number selection
        upper_bound: The upper bound for the number selection

    Returns:
        A random float number from the interval
    """
    return RNG.uniform(lower_bound, upper_bound)


def next_gaussian() -> float:
    """Returns the next pseudorandom, Gaussian ("normally") distributed
    value with mu 0.0 and sigma 1.0.

    Returns:
        The next random number
    """
    return RNG.gauss(0, 1)


_T = TypeVar("_T")  # pylint:disable=invalid-name


def choice(sequence: Sequence[_T]) -> _T:
    """Return a random element from a non-empty sequence.

    If the sequence is empty, it raises an `IndexError`.

    Args:
        sequence: The non-empty sequence to choose from

    Returns:
        An randomly selected element of the sequence
    """
    return RNG.choice(sequence)


def choices(
    population: Sequence[Any],
    weights: Sequence[float] | None = None,
    *,
    cum_weights: Sequence[float] | None = None,
    k: int = 1,
) -> list[Any]:
    """Return a k sized list of population elements chosen with replacement.

    If the relative weights or cumulative weights are not specified, the selections are
    made with equal probability.

    Args:
        population: The non-empty population to choose from
        weights: A sequence of weights
        cum_weights: A sequence of cumulative weights
        k: The size of the sample

    Returns:
        A list of sampled elements from the sequence with respect to the weight
    """
    return RNG.choices(population, weights, cum_weights=cum_weights, k=k)


def next_bool() -> bool:
    """Returns a random boolean.

    Returns:
        A random boolean
    """
    return next_float() < 0.5


def next_byte() -> int:
    """Returns a random byte.

    Returns:
        A random byte.
    """
    return RNG.getrandbits(8)


def next_bytes(length: int) -> bytes:
    """Create random bytes of given length.

    Args:
        length: the length of the bytes

    Returns:
        Random bytes of given length.
    """
    return bytes(next_byte() for _ in range(length))



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a whole-suite test generation algorithm similar to EvoSuite."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

import pynguin.configuration as config
import pynguin.coverage.branchgoals as bg
import pynguin.ga.computations as ff
import pynguin.generation.algorithms.archive as arch
from pynguin.generation.algorithms.testgenerationstrategy import TestGenerationStrategy
from pynguin.utils import randomness
from pynguin.utils.exceptions import ConstructionFailedException

if TYPE_CHECKING:
    import pynguin.ga.testsuitechromosome as tsc

# TODO(fk) instead of switching on 'use_archive' on two locations, we could
# also create another subclass?


# pylint: disable=too-few-public-methods
class WholeSuiteTestStrategy(TestGenerationStrategy[arch.CoverageArchive]):
    """Implements a whole-suite test generation algorithm similar to EvoSuite."""

    _logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self._population: list[tsc.TestSuiteChromosome] = []

    def generate_tests(
        self,
    ) -> tsc.TestSuiteChromosome:
        self.before_search_start()
        self._population = self._get_random_population()
        self._update_archive()
        self._sort_population()
        suite = self._get_solution()
        self.before_first_search_iteration(suite)
        while self.resources_left() and suite.get_fitness() != 0.0:
            self.evolve()
            suite = self._get_solution()
            self.after_search_iteration(suite)
        self.after_search_finish()
        return suite

    def evolve(self) -> None:
        """Evolve the current population and replace it with a new one."""
        new_generation = []
        new_generation.extend(self.elitism())
        while not self.is_next_population_full(new_generation):
            parent1 = self._selection_function.select(self._population, 1)[0]
            parent2 = self._selection_function.select(self._population, 1)[0]

            offspring1 = parent1.clone()
            offspring2 = parent2.clone()

            try:
                if (
                    randomness.next_float()
                    <= config.configuration.search_algorithm.crossover_rate
                ):
                    self._crossover_function.cross_over(offspring1, offspring2)

                offspring1.mutate()
                offspring2.mutate()
            except ConstructionFailedException as ex:
                self._logger.info("Crossover/Mutation failed: %s", ex)
                continue

            fitness_parents = min(parent1.get_fitness(), parent2.get_fitness())
            fitness_offspring = min(offspring1.get_fitness(), offspring2.get_fitness())
            length_parents = parent1.length() + parent2.length()
            length_offspring = offspring1.length() + offspring2.length()
            best_individual = self._get_best_individual()

            if (fitness_offspring < fitness_parents) or (
                fitness_offspring == fitness_parents
                and length_offspring <= length_parents
            ):
                for offspring in [offspring1, offspring2]:
                    if offspring.length() <= 2 * best_individual.length():
                        new_generation.append(offspring)
                    else:
                        new_generation.append(randomness.choice([parent1, parent2]))
            else:
                new_generation.append(parent1)
                new_generation.append(parent2)

        self._population = new_generation
        self._update_archive()
        self._sort_population()

    def _get_random_population(self) -> list[tsc.TestSuiteChromosome]:
        population = []
        for _ in range(config.configuration.search_algorithm.population):
            chromosome = self._chromosome_factory.get_chromosome()
            population.append(chromosome)
        return population

    def _update_archive(self) -> None:
        """Store covering test cases in archive."""
        if not config.configuration.search_algorithm.use_archive:
            return

        before = len(self._archive.uncovered_goals)
        for suite in self._population:
            self._archive.update(suite.test_case_chromosomes)
        # New goals were covered
        if before != len(self._archive.uncovered_goals):
            exclude_code: set[int] = set()
            exclude_true: set[int] = set()
            exclude_false: set[int] = set()

            # TODO(fk) Move this logic to BranchCoverageTestFitness?
            # i.e. combine with Archive.add_on_target_covered()
            for covered in self._archive.covered_goals:
                assert isinstance(covered, bg.BranchCoverageTestFitness)
                goal = covered.goal
                if goal.is_branchless_code_object:
                    exclude_code.add(goal.code_object_id)
                elif goal.is_branch:
                    branch_goal = cast(bg.BranchGoal, goal)
                    if branch_goal.value:
                        exclude_true.add(branch_goal.predicate_id)
                    else:
                        exclude_false.add(branch_goal.predicate_id)
                else:
                    raise ValueError("Unknown coverage goal")

            for func in self.test_suite_fitness_functions:
                assert isinstance(func, ff.BranchDistanceTestSuiteFitnessFunction)
                func.restrict(exclude_code, exclude_true, exclude_false)
            # Force re-computation of fitness.
            for chromosome in self._population:
                chromosome.invalidate_cache()

    def _sort_population(self) -> None:
        """Sort the population by fitness."""
        self._population.sort(key=lambda x: x.get_fitness())

    def _get_best_individual(self) -> tsc.TestSuiteChromosome:
        """Get the currently best individual.

        Returns:
            The best chromosome
        """
        return self._population[0]

    def _get_solution(self) -> tsc.TestSuiteChromosome:
        """Get the solution."""
        if config.configuration.search_algorithm.use_archive:
            # If we use an archive, use the best found solutions.
            return self.create_test_suite(self._archive.solutions)
        # If we don't use an archive, use the current best individual.
        return self._get_best_individual()

    @staticmethod
    def is_next_population_full(population: list[tsc.TestSuiteChromosome]) -> bool:
        """Check if the population is already full.

        Args:
            population: The list of chromosomes, i.e., the population

        Returns:
            Whether or not the population is already full
        """
        return len(population) >= config.configuration.search_algorithm.population

    def elitism(self) -> list[tsc.TestSuiteChromosome]:
        """Copy best individuals.

        Returns:
            A list of the best chromosomes
        """
        elite = []
        for idx in range(config.configuration.search_algorithm.elite):
            elite.append(self._population[idx].clone())
        return elite



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an implementation for a test case."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from pynguin.utils import randomness
from pynguin.utils.atomicinteger import AtomicInteger
from pynguin.utils.exceptions import ConstructionFailedException
from pynguin.utils.type_utils import is_assignable_to, is_type_unknown

if TYPE_CHECKING:
    import pynguin.assertion.assertion as ass
    import pynguin.testcase.statement as stmt
    import pynguin.testcase.testcasevisitor as tcv
    import pynguin.testcase.variablereference as vr


# pylint: disable=too-many-public-methods
class TestCase(metaclass=ABCMeta):
    """An abstract base implementation for a test case.

    Serves as an interface for test-case implementations
    """

    _id_generator = AtomicInteger()

    def __init__(self) -> None:
        self._statements: list[stmt.Statement] = []

    @property
    def statements(self) -> list[stmt.Statement]:
        """Provides the list of statements in this test case.

        Returns:
            The list of statements in this test case
        """
        return self._statements

    @abstractmethod
    def accept(self, visitor: tcv.TestCaseVisitor) -> None:
        """Handles a test visitor.

        Args:
            visitor: The test visitor to accept
        """

    @abstractmethod
    def add_statement(
        self, statement: stmt.Statement, position: int = -1
    ) -> vr.VariableReference | None:
        """Adds a new statement to the test case.

        The optional position parameter specifies the position.  If it is not given,
        the statement will be added to the end of the test case.

        Args:
            statement: The new statement
            position: The optional position where to put the statement

        Returns:  # noqa: DAR202
            The return value of the statement.  Notice that the test might
            choose to modify the statement you inserted.  You should use the returned
            variable reference and not use references. Can be None, if this statement
            does not create a variable.
        """

    @abstractmethod
    def add_variable_creating_statement(
        self, statement: stmt.VariableCreatingStatement, position: int = -1
    ) -> vr.VariableReference:
        """Overloaded version of add_statement that adds a statement
        which creates a variable.

        Args:
            statement: The new statement
            position: The optional position where to put the statement

        Returns:  # noqa: DAR202
            The return value of the statement.  Notice that the test might
            choose to modify the statement you inserted.  You should use the returned
            variable reference and not use references.
        """

    @abstractmethod
    def add_statements(self, statements: list[stmt.Statement]) -> None:
        """Adds a list of statements to the end of the test case.

        Args:
            statements: The list of statements to add
        """

    @abstractmethod
    def append_test_case(self, test_case: TestCase) -> None:
        """Appends a test case to this test case.

        Args:
            test_case: The test case to append
        """

    @abstractmethod
    def remove(self, position: int) -> None:
        """Removes a statement a the given position

        Args:
            position: The position of the test case to be removed
        """

    @abstractmethod
    def remove_statement(self, statement: stmt.Statement) -> None:
        """Remove the given statement from this test case.

        Args:
            statement: The statement to remove.
        """

    @abstractmethod
    def chop(self, pos: int) -> None:
        """Remove all statements after a given position.

        Args:
            pos: The length of the test case after chopping
        """

    @abstractmethod
    def contains(self, statement: stmt.Statement) -> bool:
        """Determines whether or not the test case contains a specific statement.

        Args:
            statement: The statement to search in the test case

        Returns:
            Whether or not the test case contains the statement  # noqa: DAR202
        """

    @abstractmethod
    def get_statement(self, position: int) -> stmt.Statement:
        """Provides access to a statement at a given position.

        Args:
            position: The position of the statement in the test case

        Returns:
            The statement at the position  # noqa: DAR202
        """

    @abstractmethod
    def set_statement(
        self, statement: stmt.Statement, position: int
    ) -> vr.VariableReference | None:
        """Set new statement at position.

        Args:
            statement: the new statement
            position: the position for the new statement

        Returns:
            A variable reference to the statements return value, if any  # noqa: DAR202
        """

    @abstractmethod
    def has_statement(self, position: int) -> bool:
        """Check if there is a statement at the given position.

        Args:
            position: The index of the statement

        Returns:
            Whether or not there is a statement at the given position  # noqa: DAR202
        """

    @abstractmethod
    def clone(self, limit: int | None = None) -> TestCase:
        """Provides a deep copy of the test case.

        Args:
            limit: Clone this test case only upto the given number of statements.

        Returns:
            A deep copy of this test case  # noqa: DAR202
        """

    @abstractmethod
    def size(self) -> int:
        """Provides the number of statements in the test case.

        Returns:
            The number of statements in the test case  # noqa: DAR202
        """

    @abstractmethod
    def size_with_assertions(self) -> int:
        """Provides the number of statements and assertions in the test case.

        Returns:
            The number of statements and assertions in the test case # noqa: DAR202
        """

    @abstractmethod
    def get_assertions(self) -> list[ass.Assertion]:
        """Get all assertions that exist for this test case."""

    @abstractmethod
    def get_dependencies(self, var: vr.VariableReference) -> set[vr.VariableReference]:
        """Provides all variables on which var depends.

        Args:
            var: the variable whose dependencies we are looking for.

        Returns:
            a set of variables on which var depends on. # noqa: DAR202
        """

    def get_objects(
        self, parameter_type: type | None, position: int
    ) -> list[vr.VariableReference]:
        """Provides a list of variable references satisfying a certain type before a
        given position.

        If the position value is larger than the number of statements, only these
        statements will be considered.  Otherwise the first `position` statements of
        the test case will be considered.

        If the type for which we search is not specified, all objects up to the given
        position are returned.

        Args:
            parameter_type: The type of the parameter we search references for
            position: The position in the statement list until we search

        Returns:
            A list of variable references satisfying the parameter type
        """
        if is_type_unknown(parameter_type):
            return self.get_all_objects(position)

        variables: list[vr.VariableReference] = []
        bound = min(len(self._statements), position)
        for i in range(bound):
            statement = self._statements[i]
            var = statement.ret_val
            if var is None:
                continue
            if not var.is_none_type() and is_assignable_to(var.type, parameter_type):
                variables.append(var)

        return variables

    def get_all_objects(self, position: int) -> list[vr.VariableReference]:
        """Get all objects that are defined up to the given position (exclusive).

        Args:
            position: the position

        Returns:
            A list of all objects defined up to the given position
        """
        variables: list[vr.VariableReference] = []
        bound = min(len(self._statements), position)
        for i in range(bound):
            var = self.get_statement(i).ret_val
            if var is None:
                continue
            if not var.is_none_type():
                variables.append(var)
        return variables

    def get_random_object(
        self, parameter_type: type | None, position: int
    ) -> vr.VariableReference:
        """Get a random object of the given type up to the given position (exclusive).

        Args:
            parameter_type: the parameter type
            position: the position

        Returns:
            A random object of given type up to the given position

        Raises:
            ConstructionFailedException: if no object could be found
        """
        variables = self.get_objects(parameter_type, position)
        if len(variables) == 0:
            raise ConstructionFailedException(
                f"Found no variables of type {parameter_type} at position {position}"
            )
        return randomness.choice(variables)
